# Copyright (C) 2016-2017 Cumulus Networks, inc.
#
# All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement available
# at the following locations:.
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
#
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

"""
Usage:
    # L3 clear commands
    frr clear bgp [vrf <text>] (<bgppeer>|all) [soft in|soft out]
    frr clear bgp [vrf <text>] prefix <ipv4/prefixlen>
    frr clear ospf interface [<interface>]

    # L3 show commands
    frr show debugs
    frr show bgp [vrf <text>] [summary] [json]
    frr show bgp [vrf <text>] (<ipv4>|<ipv4/prefixlen>|<ipv6>|<ipv6/prefixlen>) [bestpath|multipath] [json]
    frr show bgp ipv4 (unicast|labeled-unicast) [<ipv4>|<ipv4/prefixlen>|summary] [json]
    frr show bgp ipv6 (unicast|labeled-unicast) [<ipv6>|<ipv6/prefixlen>|summary] [json]
    frr show bgp [vrf <text>] neighbor [<bgppeer>] [json]
    frr show bgp [vrf <text>] neighbor <bgppeer> (advertised-routes|received-routes|routes) [json]
    frr show bgp [vrf <text>] (ipv4|ipv6) (unicast|labeled-unicast) neighbor <bgppeer> (advertised-routes|received-routes|routes) [json]
    frr show bgp [vrf <text>] update-groups
    frr show bgp [l2vpn] evpn import-rt
    frr show bgp [l2vpn] evpn route
    frr show bgp [l2vpn] evpn route rd <rd>
    frr show bgp [l2vpn] evpn route rd <rd> mac <mac> [ip <ip>]
    frr show bgp [l2vpn] evpn route rd <rd> type (macip|multicast)
    frr show bgp [l2vpn] evpn route type (macip|multicast)
    frr show bgp [l2vpn] evpn route vni all
    frr show bgp [l2vpn] evpn route vni <1-16777215>
    frr show bgp [l2vpn] evpn route vni <1-16777215> vtep <ipv4>
    frr show bgp [l2vpn] evpn route vni all vtep <ipv4>
    frr show bgp [l2vpn] evpn route vni <1-16777215> mac <mac> [ip <ip>]
    frr show bgp [l2vpn] evpn route vni <1-16777215> multicast <ipv4>
    frr show bgp [l2vpn] evpn route vni <1-16777215> type (macip|multicast)
    frr show bgp [l2vpn] evpn summary [json]
    frr show bgp [l2vpn] evpn vni [<1-16777215>]
    frr show evpn arp-cache vni all [json]
    frr show evpn arp-cache vni <1-16777215> [json]
    frr show evpn arp-cache vni <1-16777215> ip <ip> [json]
    frr show evpn arp-cache vni <1-16777215> vtep <ipv4> [json]
    frr show evpn mac vni all [json]
    frr show evpn mac vni <1-16777215> [json]
    frr show evpn mac vni <1-16777215> mac <mac>
    frr show evpn mac vni <1-16777215> vtep <ipv4> [json]
    frr show evpn mac vni all vtep <ipv4> [json]
    frr show evpn vni [<1-16777215>]  [json]
    frr show ospf [json]
    frr show ospf database (asbr-summary|external|network|nssa-external|opaque-area|opaque-as|opaque-link|router|summary) [<ipv4>|adv-router <ipv4-adv-router>|self-originate]
    frr show ospf database [max-age|self-originate]
    frr show ospf interface [<interface>] [json]
    frr show ospf neighbor [all|<interface>|<ipv4>|detail] [json]
    frr show ospf route
    frr show ospf6
    frr show ospf6 database [adv-router <ipv4>|linkstate-id <ipv4-id>|self-originated] [detail|dump|internal]
    frr show ospf6 interface [<interface>]
    frr show ospf6 linkstate [detail|router <ipv4>]
    frr show ospf6 neighbor [detail|drchoice]
    frr show ospf6 redistribute
    frr show ospf6 route [<ipv6>|<ipv6/prefixlen>|detail|summary]
    frr show ospf6 route (intra-area|inter-area|external-1|external-2) [detail]
    frr show ospf6 route <ipv6/prefixlen> longer [detail]
    frr show ospf6 route <ipv6/prefixlen> match [detail]
    frr show ospf6 spf tree
    frr show route [vrf <text>] [<ipv4>|<ipv4/prefixlen>|<ipv6>|<ipv6/prefixlen>|bgp|connected|json|kernel|ospf|ospf6|pim|rip|static|summary|supernets-only|table] [json]
    frr show route [vrf <text>] (<ipv4/prefixlen>|<ipv6/prefixlen>) longer-prefixes [json]
    frr show ip nht [vrf <text>|route-map]

    # L3 global configuration
    frr (add|del) routing protocol (bgp|ospf|static|table) route-map <route-map>
    #frr (add|del) routing import-table <1-252>
    #frr (add|del) routing import-table <1-252> distance <1-255>
    frr (add|del) routing import-table <1-252> [distance <1-255>] [route-map <route-map>]
    frr add routing route <ipv4/prefixlen> (<ipv4>|<interface>|Null0|blackhole|reject) [<1-255>] [vrf <text>]
    frr add routing route <ipv6/prefixlen> (<ipv6>|<interface>|Null0|blackhole|reject) [<1-255>] [vrf <text>]
    frr del routing route <ipv4/prefixlen> (<ipv4>|<interface>|Null0) [<1-255>] [vrf <text>]
    frr del routing route <ipv6/prefixlen> (<ipv6>|<interface>|Null0) [<1-255>] [vrf <text>]
    frr add routing route <ipv4/prefixlen> (<ipv4>|<interface>|Null0|blackhole|reject) [<1-255>] label <text>
    frr add routing route <ipv6/prefixlen> (<ipv6>|<interface>|Null0|blackhole|reject) [<1-255>] label <text>
    frr del routing route <ipv4/prefixlen> (<ipv4>|<interface>|Null0) [<1-255>] label <text>
    frr del routing route <ipv6/prefixlen> (<ipv6>|<interface>|Null0) [<1-255>] label <text>
    frr (add|del) routing log file <text> [alerts|critical|debugging|emergencies|errors|informational|notifications|warnings]
    frr (add|del) routing log syslog [alerts|critical|debugging|emergencies|errors|informational|notifications|warnings]
    frr (add|del) routing log timestamp precision <0-6>
    frr (add|del) routing zebra debug (events|fpm|kernel|mpls|nht|vxlan)
    frr (add|del) routing zebra debug packet [recv|send] [detail]
    frr (add|del) routing zebra debug rib [detailed]
    frr (add|del) routing mroute debug [detail]
    frr (add|del) routing ptm-enable
    frr add       routing password <text>
    frr (add|del) routing enable password <text>
    frr add       routing line vty
    frr (add|del) routing line vty exec-timeout <0-35791> <0-2147483>
    frr add       routing service integrated-vtysh-config

    frr (add|del) (bond <interface>|interface <interface>|vlan <number-range-list>) ptm-enable
    frr add       (bond <interface>|interface <interface>|vlan <number-range-list>) ipv6 nd ra-interval (<1-1800>|msec <70-1800000>)
    frr del       (bond <interface>|interface <interface>|vlan <number-range-list>) ipv6 nd ra-interval [<1-1800>]
    frr del       (bond <interface>|interface <interface>|vlan <number-range-list>) ipv6 nd ra-interval msec [<70-1800000>]
    frr add       (bond <interface>|interface <interface>|vlan <number-range-list>) ipv6 nd ra-lifetime <0-9000>
    frr del       (bond <interface>|interface <interface>|vlan <number-range-list>) ipv6 nd ra-lifetime [<0-9000>]
    frr (add|del) (bond <interface>|interface <interface>|vlan <number-range-list>) ipv6 nd suppress-ra
    frr (add|del) (bond <interface>|interface <interface>|vlan <number-range-list>) ipv6 nd prefix <ipv6/prefixlen> [router-address|off-link|off-link no-autoconfig]
    frr (add|del) (bond <interface>|interface <interface>|vlan <number-range-list>) ipv6 nd prefix <ipv6/prefixlen> (<0-4294967295>|infinite) (<0-4294967295>|infinite) [router-address|off-link|off-link no-autoconfig]

    # L3 BGP 'global' configuration
    frr (add|del) bgp [vrf <text>] autonomous-system <1-4294967295>
    frr (add|del) bgp [vrf <text>] always-compare-med
    frr (add|del) bgp [vrf <text>] bestpath as-path multipath-relax [as-set|no-as-set]
    frr (add|del) bgp [vrf <text>] bestpath compare-routerid
    frr (add|del) bgp [vrf <text>] bestpath med missing-as-worst
    frr (add|del) bgp [vrf <text>] cluster-id (<ipv4>|<1-4294967295>)
    frr add       bgp debug bestpath <ip/prefixlen>
    frr del       bgp debug bestpath [<ip/prefixlen>]
    frr add       bgp debug keepalives (<ip>|<interface>)
    frr del       bgp debug keepalives [<ip>|<interface>]
    frr add       bgp debug neighbor-events [<ip>|<interface>]
    frr del       bgp debug neighbor-events [<ip>|<interface>]
    frr (add|del) bgp debug updates [in|out]
    frr add       bgp debug updates prefix <ip/prefixlen>
    frr del       bgp debug updates prefix [<ip/prefixlen>]
    frr (add|del) bgp debug zebra
    frr add       bgp debug zebra prefix <ip/prefixlen>
    frr del       bgp debug zebra prefix [<ip/prefixlen>]
    frr (add|del) bgp [vrf <text>] default ipv4-unicast
    frr (add|del) bgp [vrf <text>] listen limit <1-5000>
    frr (add|del) bgp [vrf <text>] listen range (<ipv4/prefixlen>|<ipv6/prefixlen>) peer-group <peergroup>
    frr add       bgp [vrf <text>] neighbor <bgppeer> advertisement-interval <0-600>
    frr del       bgp [vrf <text>] neighbor <bgppeer> advertisement-interval [<0-600>]
    frr (add|del) bgp [vrf <text>] neighbor <bgppeer> bfd [<2-255> <50-60000> <50-60000>]
    frr (add|del) bgp [vrf <text>] neighbor <bgppeer> bfd [multihop]
    frr (add|del) bgp [vrf <text>] neighbor <bgppeer> capability extended-nexthop
    frr add       bgp [vrf <text>] neighbor <bgppeer> description <wildcard>
    frr del       bgp [vrf <text>] neighbor <bgppeer> description [<wildcard>]
    frr (add|del) bgp [vrf <text>] neighbor <bgppeer> ebgp-multihop [<1-255>]
    frr add       bgp [vrf <text>] neighbor <bgppeer> password <text-password>
    frr del       bgp [vrf <text>] neighbor <bgppeer> password [<text-password>]
    frr (add|del) bgp [vrf <text>] neighbor (<ip>|<bgppeer>|<interface>) peer-group <peergroup>
    frr (add|del) bgp [vrf <text>] neighbor (<ip>|<bgppeer>|<interface>) remote-as (<1-4294967295>|external|internal)
    frr (add|del) bgp [vrf <text>] neighbor <bgppeer> shutdown
    frr add       bgp [vrf <text>] neighbor <bgppeer> timers <0-65535> <3-65535>
    frr del       bgp [vrf <text>] neighbor <bgppeer> timers [<0-65535>] [<3-65535>]
    frr add       bgp [vrf <text>] neighbor <bgppeer> timers connect <1-65535>
    frr del       bgp [vrf <text>] neighbor <bgppeer> timers connect [<1-65535>]
    frr (add|del) bgp [vrf <text>] neighbor <bgppeer> ttl-security hops <1-254>
    frr add       bgp [vrf <text>] neighbor <bgppeer> update-source (<ipv4>|<ipv6>|<interface-source>)
    frr del       bgp [vrf <text>] neighbor <bgppeer> update-source [<ipv4>|<ipv6>|<interface-source>]
    frr (add|del) bgp [vrf <text>] neighbor <interface> interface v6only
    frr (add|del) bgp [vrf <text>] neighbor <interface> interface [v6only] peer-group <peergroup>
    frr (add|del) bgp [vrf <text>] neighbor <interface> interface remote-as (<1-4294967295>|external|internal)
    frr (add|del) bgp [vrf <text>] neighbor <text-peer-group> peer-group
    frr (add|del) bgp [vrf <text>] route-map delay-timer <0-600>
    frr add       bgp [vrf <text>] router-id <ipv4>
    frr del       bgp [vrf <text>] router-id [<ipv4>]
    frr (add|del) bgp [vrf <text>] timers <0-65535> <0-65535>
    frr add       bgp [vrf <text>] update-delay <0-3600> [<1-3600>]
    frr del       bgp [vrf <text>] update-delay <0-3600> <1-3600>
    frr del       bgp [vrf <text>] update-delay [<0-3600>]

    # L3 BGP per address-family configuration
    frr (add|del) bgp [vrf <text>] [ipv4 unicast] aggregate-address <ipv4/prefixlen> [as-set] [summary-only]
    frr (add|del) bgp [vrf <text>] ipv6 unicast aggregate-address <ipv6/prefixlen> [summary-only]
    frr (add|del) bgp [vrf <text>] network (<ipv4/prefixlen>|<ipv6/prefixlen>) [label-index <0-1048560>] [route-map <text-route-map>]
    frr (add|del) bgp [vrf <text>] ipv4 unicast network <ipv4/prefixlen> [label-index <0-1048560>] [route-map <text-route-map>]
    frr (add|del) bgp [vrf <text>] ipv6 unicast network <ipv6/prefixlen> [label-index <0-1048560>] [route-map <text-route-map>]
    frr (add|del) bgp [vrf <text>] [ipv4 unicast] redistribute (connected|kernel|ospf|pim|static|table|table <1-65535>)
    frr (add|del) bgp [vrf <text>] [ipv4 unicast] redistribute (connected|kernel|ospf|pim|static|table|table <1-65535>) metric <0-4294967295>
    frr (add|del) bgp [vrf <text>] [ipv4 unicast] redistribute (connected|kernel|ospf|pim|static|table|table <1-65535>) route-map <route-map>
    frr (add|del) bgp [vrf <text>] [ipv4 unicast] redistribute (connected|kernel|ospf|pim|static|table|table <1-65535>) metric <0-4294967295> route-map <route-map>
    frr (add|del) bgp [vrf <text>] ipv6 unicast redistribute (connected|kernel|ospf6|static|table)
    frr (add|del) bgp [vrf <text>] ipv6 unicast redistribute (connected|kernel|ospf6|static|table) metric <0-4294967295>
    frr (add|del) bgp [vrf <text>] ipv6 unicast redistribute (connected|kernel|ospf6|static|table) route-map <route-map>
    frr (add|del) bgp [vrf <text>] ipv6 unicast redistribute (connected|kernel|ospf6|static|table) metric <0-4294967295> route-map <route-map>
    frr add       bgp [vrf <text>] [ipv4 unicast|ipv6 unicast] maximum-paths [ibgp] <1-256>
    frr del       bgp [vrf <text>] [ipv4 unicast|ipv6 unicast] maximum-paths [ibgp] [<1-256>]
    frr (add|del) bgp [vrf <text>] [ipv4 unicast|ipv4 labeled-unicast|ipv6 unicast|ipv6 labeled-unicast] neighbor <bgppeer> activate
    frr (add|del) bgp [vrf <text>] [ipv4 unicast|ipv4 labeled-unicast|ipv6 unicast|ipv6 labeled-unicast] neighbor <bgppeer> addpath-tx-all-paths
    frr (add|del) bgp [vrf <text>] [ipv4 unicast|ipv4 labeled-unicast|ipv6 unicast|ipv6 labeled-unicast] neighbor <bgppeer> addpath-tx-bestpath-per-AS
    frr add       bgp [vrf <text>] [ipv4 unicast|ipv4 labeled-unicast|ipv6 unicast|ipv6 labeled-unicast] neighbor <bgppeer> allowas-in (<1-10>|origin)
    frr del       bgp [vrf <text>] [ipv4 unicast|ipv4 labeled-unicast|ipv6 unicast|ipv6 labeled-unicast] neighbor <bgppeer> allowas-in [<1-10>|origin]
    frr (add|del) bgp [vrf <text>] [ipv4 unicast|ipv4 labeled-unicast|ipv6 unicast|ipv6 labeled-unicast] neighbor <bgppeer> default-originate
    frr (add|del) bgp [vrf <text>] [ipv4 unicast|ipv4 labeled-unicast|ipv6 unicast|ipv6 labeled-unicast] neighbor <bgppeer> default-originate route-map <route-map>
    frr (add|del) bgp [vrf <text>] [ipv4 unicast|ipv4 labeled-unicast|ipv6 unicast|ipv6 labeled-unicast] neighbor <bgppeer> filter-list <as-path-access-list> (in|out)
    frr (add|del) bgp [vrf <text>] [ipv4 unicast|ipv4 labeled-unicast|ipv6 unicast|ipv6 labeled-unicast] neighbor <bgppeer> next-hop-self [force]
    frr (add|del) bgp [vrf <text>] [ipv4 unicast|ipv4 labeled-unicast] neighbor <bgppeer> prefix-list <prefix-list-v4> (in|out)
    frr (add|del) bgp [vrf <text>] ipv6 unicast neighbor <bgppeer> prefix-list <prefix-list-v6> (in|out)
    frr (add|del) bgp [vrf <text>] [ipv4 unicast|ipv4 labeled-unicast|ipv6 unicast|ipv6 labeled-unicast] neighbor <bgppeer> remove-private-AS [all] [replace-AS]
    frr (add|del) bgp [vrf <text>] [ipv4 unicast|ipv4 labeled-unicast|ipv6 unicast|ipv6 labeled-unicast] neighbor <bgppeer> route-map <route-map> (in|out)
    frr (add|del) bgp [vrf <text>] [ipv4 unicast|ipv4 labeled-unicast|ipv6 unicast|ipv6 labeled-unicast] neighbor <bgppeer> route-reflector-client
    frr (add|del) bgp [vrf <text>] [ipv4 unicast|ipv4 labeled-unicast|ipv6 unicast|ipv6 labeled-unicast] neighbor <bgppeer> send-community [both|extended|standard]
    frr (add|del) bgp [vrf <text>] [ipv4 unicast|ipv4 labeled-unicast|ipv6 unicast|ipv6 labeled-unicast] neighbor <bgppeer> soft-reconfiguration inbound
    frr add       bgp [vrf <text>] [ipv4 unicast|ipv6 unicast] table-map <route-map>
    frr del       bgp [vrf <text>] [ipv4 unicast|ipv6 unicast] table-map [<route-map>]

    # EVPN
    frr (add|del) bgp [l2vpn] evpn advertise-all-vni
    frr (add|del) bgp [l2vpn] evpn advertise-default-gw
    frr (add|del) bgp [l2vpn] evpn neighbor <bgppeer> activate
    frr add       bgp [l2vpn] evpn neighbor <bgppeer> allowas-in <1-10>
    frr del       bgp [l2vpn] evpn neighbor <bgppeer> allowas-in [<1-10>]
    frr (add|del) bgp [l2vpn] evpn neighbor <bgppeer> route-reflector-client
    frr (add|del) bgp [l2vpn] evpn vni <1-16777215>
    frr (add|del) bgp [l2vpn] evpn vni <1-16777215> rd <rd>
    frr (add|del) bgp [l2vpn] evpn vni <1-16777215> route-target (both|export|import) <route-target>
    frr (add|del) bgp [l2vpn] evpn vni <1-16777215> advertise-default-gw

    # L3 OSPF configuration
    frr del ospf
    frr (add|del) ospf area (<ipv4>|<0-4294967295>) stub [no-summary]
    frr (add|del) ospf debug ism [status|events|timers]
    frr (add|del) ospf debug lsa [generate|flooding|install|refresh]
    frr (add|del) ospf debug nsm [status|events|timers]
    frr (add|del) ospf debug packet (hello|dd|ls-request|ls-update|ls-ack|all) [send|recv|detail]
    frr (add|del) ospf debug zebra
    frr (add|del) ospf default-information originate [always]
    frr (add|del) ospf log-adjacency-changes [detail]
    frr (add|del) ospf network <ipv4/prefixlen> area (<ipv4>|<0-4294967295>)
    frr del       ospf network <ipv4/prefixlen>
    frr (add|del) ospf passive-interface (default|<interface>)
    frr (add|del) ospf redistribute (bgp|connected|kernel|static|table|table <1-65535>)
    frr (add|del) ospf redistribute (bgp|connected|kernel|static|table|table <1-65535>) metric <0-16777214>
    frr (add|del) ospf redistribute (bgp|connected|kernel|static|table|table <1-65535>) metric-type (1|2)
    frr (add|del) ospf redistribute (bgp|connected|kernel|static|table|table <1-65535>) route-map <route-map>
    frr add       ospf router-id <ipv4>
    frr del       ospf router-id [<ipv4>]
    frr add       ospf timers lsa min-arrival <0-600000>
    frr del       ospf timers lsa min-arrival [<0-600000>]
    frr (add|del) ospf timers throttle spf <0-600000> <0-600000> <0-600000>
    frr add       (bond <interface>|interface <interface>|vlan <number-range-list>|loopback lo) ospf area (<ipv4>|<0-4294967295>)
    frr del       (bond <interface>|interface <interface>|vlan <number-range-list>|loopback lo) ospf area [<ipv4>|<0-4294967295>]
    frr (add|del) (bond <interface>|interface <interface>|vlan <number-range-list>) ospf authentication message-digest
    frr (add|del) (bond <interface>|interface <interface>|vlan <number-range-list>) ospf cost <1-65535>
    frr (add|del) (bond <interface>|interface <interface>|vlan <number-range-list>) ospf message-digest-key <1-255> md5 <text>
    frr (add|del) (bond <interface>|interface <interface>|vlan <number-range-list>) ospf bfd [<2-255> <50-60000> <50-60000>]
    frr add       (bond <interface>|interface <interface>|vlan <number-range-list>) ospf dead-interval <1-65535>
    frr del       (bond <interface>|interface <interface>|vlan <number-range-list>) ospf dead-interval [<1-65535>]
    frr add       (bond <interface>|interface <interface>|vlan <number-range-list>) ospf hello-interval <1-65535>
    frr del       (bond <interface>|interface <interface>|vlan <number-range-list>) ospf hello-interval [<1-65535>]
    frr add       (bond <interface>|interface <interface>|vlan <number-range-list>) ospf network (broadcast|non-broadcast|point-to-multipoint|point-to-point)
    frr del       (bond <interface>|interface <interface>|vlan <number-range-list>) ospf network [broadcast|non-broadcast|point-to-multipoint|point-to-point]

    # L3 OSPFv3 configuration
    frr del ospf6
    frr (add|del) ospf6 area <ipv4> stub [no-summary]
    frr (add|del) ospf6 auto-cost reference-bandwidth <1-4294967>
    frr (add|del) ospf6 debug (abr|asbr|flooding|interface)
    frr (add|del) ospf6 debug border-routers [area-id <ipv4>|router-id <ipv4>]
    frr (add|del) ospf6 debug lsa (router|network|inter-prefix|inter-router|as-external|link|intra-prefix|unknown) [originate|examine|flooding]
    frr (add|del) ospf6 debug message (unknown|hello|dbdesc|lsreq|lsupdate|lsack|all) [send|recv]
    frr (add|del) ospf6 debug neighbor [state|event]
    frr (add|del) ospf6 debug route (table|intra-area|inter-area|memory)
    frr (add|del) ospf6 debug spf (database|process|time)
    frr (add|del) ospf6 debug zebra [send|recv]
    frr (add|del) ospf6 interface <interface> area <ipv4>
    frr (add|del) ospf6 log-adjacency-changes [detail]
    frr (add|del) ospf6 redistribute (bgp|connected|kernel|static|table) [route-map <route-map>]
    frr add       ospf6 router-id <ipv4>
    frr del       ospf6 router-id [<ipv4>]
    frr (add|del) ospf6 stub-router administrative
    frr add       ospf6 timers lsa min-arrival <0-600000>
    frr del       ospf6 timers lsa min-arrival [<0-600000>]
    frr (add|del) ospf6 timers throttle spf <0-600000> <0-600000> <0-600000>
    frr (add|del) (bond <interface>|interface <interface>|vlan <number-range-list>) ospf6 bfd [<2-255> <50-60000> <50-60000>]
    frr add       (bond <interface>|interface <interface>|vlan <number-range-list>|loopback lo) ospf6 cost <1-65535>
    frr del       (bond <interface>|interface <interface>|vlan <number-range-list>|loopback lo) ospf6 cost [<1-65535>]
    frr add       (bond <interface>|interface <interface>|vlan <number-range-list>) ospf6 dead-interval <1-65535>
    frr del       (bond <interface>|interface <interface>|vlan <number-range-list>) ospf6 dead-interval [<1-65535>]
    frr add       (bond <interface>|interface <interface>|vlan <number-range-list>) ospf6 hello-interval <1-65535>
    frr del       (bond <interface>|interface <interface>|vlan <number-range-list>) ospf6 hello-interval [<1-65535>]
    frr add       (bond <interface>|interface <interface>|vlan <number-range-list>) ospf6 ifmtu <1-65535>
    frr del       (bond <interface>|interface <interface>|vlan <number-range-list>) ospf6 ifmtu [<1-65535>]
    frr add       (bond <interface>|interface <interface>|vlan <number-range-list>|loopback lo) ospf6 instance-id <0-255>
    frr del       (bond <interface>|interface <interface>|vlan <number-range-list>|loopback lo) ospf6 instance-id [<0-255>]
    frr (add|del) (bond <interface>|interface <interface>|vlan <number-range-list>) ospf6 mtu-ignore
    frr add       (bond <interface>|interface <interface>|vlan <number-range-list>) ospf6 network (broadcast|point-to-point)
    frr del       (bond <interface>|interface <interface>|vlan <number-range-list>) ospf6 network [broadcast|point-to-point]
    frr (add|del) (bond <interface>|interface <interface>|vlan <number-range-list>|loopback lo) ospf6 passive
    frr add       (bond <interface>|interface <interface>|vlan <number-range-list>) ospf6 priority <0-255>
    frr del       (bond <interface>|interface <interface>|vlan <number-range-list>) ospf6 priority [<0-255>]
    frr add       (bond <interface>|interface <interface>|vlan <number-range-list>) ospf6 retransmit-interval <1-65535>
    frr del       (bond <interface>|interface <interface>|vlan <number-range-list>) ospf6 retransmit-interval [<1-65535>]
    frr add       (bond <interface>|interface <interface>|vlan <number-range-list>) ospf6 transmit-delay <1-3600>
    frr del       (bond <interface>|interface <interface>|vlan <number-range-list>) ospf6 transmit-delay [<1-3600>]

    # L3 multicast debugs
    frr (add|del) msdp debug (events|packets)
    frr (add|del) igmp debug (events|packets|trace)
    frr (add|del) pim debug packets [hello|joins|register]
    frr add pim debug [events|trace|zebra]
    frr del pim debug (events|trace|zebra)

    # L3 multicast show commands
    frr show igmp [vrf <text>] (join|sources)
    frr show igmp [vrf <text>] groups [json]
    frr show igmp [vrf <text>] interface [detail|<interface>] [json]
    frr show mroute [vrf <text>] [count|json]
    frr show msdp [vrf <text>] mesh-group [json]
    frr show msdp [vrf <text>] peer [detail|<ipv4>] [json]
    frr show msdp [vrf <text>] sa [detail] [json]
    frr show msdp [vrf <text>] sa <ipv4> [json]
    frr show msdp [vrf <text>] sa <ipv4> <ipv4-mcast-group> [json]
    frr show pim [vrf <text>] (assert|assert-internal|assert-metric|assert-winner-metric|secondary)
    frr show pim [vrf <text>] (join|local-membership|rpf|rp-info|upstream|upstream-join-desired|upstream-rpf) [json]
    frr show pim [vrf <text>] group-type [json]
    frr show pim [vrf <text>] group-type <ipv4-mcast-group> [json]
    frr show pim [vrf <text>] interface [detail|<interface>] [json]
    frr show pim [vrf <text>] interface traffic [<interface>] [json]
    frr show pim [vrf <text>] neighbor [detail|<interface>|<ipv4>] [json]
    frr show pim [vrf <text>] nexthop-lookup <ipv4-source> <ipv4-mcast-group>
    frr show pim [vrf <text>] state [json]
    frr show pim [vrf <text>] state <ipv4-source> [json]
    frr show pim [vrf <text>] state <ipv4-source> <ipv4-mcast-group> [json]

    # L3 multicast configuration
    frr (add|del) pim [vrf <text>] rp <ipv4> [<ipv4/prefixlen>]
    frr add       pim [vrf <text>] join-prune-interval <60-600>
    frr del       pim [vrf <text>] join-prune-interval [<60-600>]
    frr add       pim [vrf <text>] keep-alive-timer <31-60000>
    frr del       pim [vrf <text>] keep-alive-timer [<31-60000>]
    frr add       pim [vrf <text>] packets <1-100>
    frr del       pim [vrf <text>] packets [<1-100>]
    frr add       pim [vrf <text>] register-suppress-time <5-60000>
    frr del       pim [vrf <text>] register-suppress-time [<5-60000>]
    frr add       pim [vrf <text>] ssm prefix-list <prefix-list-v4>
    frr del       pim [vrf <text>] ssm prefix-list [<prefix-list-v4>]
    frr (add|del) pim [vrf <text>] spt-switchover infinity-and-beyond [prefix-list <prefix-list-v4>]
    frr (add|del) pim [vrf <text>] ecmp [rebalance]
    frr (add|del) (bond <interface>|interface <interface>|vlan <number-range-list>) igmp
    frr add       (bond <interface>|interface <interface>|vlan <number-range-list>) igmp version (2|3)
    frr del       (bond <interface>|interface <interface>|vlan <number-range-list>) igmp version [2]
    frr (add|del) (bond <interface>|interface <interface>|vlan <number-range-list>) igmp join <ipv4-mcast-group> <ipv4-source>
    frr add       (bond <interface>|interface <interface>|vlan <number-range-list>) igmp query-interval <1-1800>
    frr del       (bond <interface>|interface <interface>|vlan <number-range-list>) igmp query-interval [<1-1800>]
    frr add       (bond <interface>|interface <interface>|vlan <number-range-list>) igmp query-max-response-time <10-250>
    frr del       (bond <interface>|interface <interface>|vlan <number-range-list>) igmp query-max-response-time [<10-250>]
    frr add       (bond <interface>|interface <interface>|vlan <number-range-list>) pim drpriority <1-4294967295>
    frr del       (bond <interface>|interface <interface>|vlan <number-range-list>) pim drpriority [<1-4294967295>]
    frr add       (bond <interface>|interface <interface>|vlan <number-range-list>) pim hello <1-180> <3-180>
    frr del       (bond <interface>|interface <interface>|vlan <number-range-list>) pim hello <1-180> <3-180>
    frr del       (bond <interface>|interface <interface>|vlan <number-range-list>) pim hello [<1-180>]
    frr (add|del) (bond <interface>|interface <interface>|vlan <number-range-list>) pim bfd [<2-255> <50-60000> <50-60000>]
    frr (add|del) (bond <interface>|interface <interface>|vlan <number-range-list>|loopback lo) pim sm
    frr (add|del) (bond <interface>|interface <interface>|vlan <number-range-list>|loopback lo) pim use-source <ipv4>
    frr (add|del) msdp mesh-group <text> (member|source) <ipv4>
    frr (add|del) routing mroute <ipv4/prefixlen> (<ipv4>|<interface>) [<1-255>]

    # SNMP Configuration
    frr add routing agentx

    # SNMP Configuration
    frr add routing agentx

    # as-path lists
    frr (add|del) routing as-path access-list <text> (deny|permit) <wildcard>
    frr show ip as-path-access-list [<as-path-access-list>]

    # community-list
    frr (add|del) routing community-list standard <text> (deny|permit) (local-AS|no-advertise|no-export|internet|<community>)
    frr (add|del) routing community-list expanded <text> (deny|permit) <regex>
    frr (add|del) routing extcommunity-list standard <text> (deny|permit) (rt|soo) <extcommunity>
    frr (add|del) routing extcommunity-list expanded <text> (deny|permit) <regex>

    frr show ip community-list <community-list>
    frr show ip extcommunity-list <extcommunity-list>

    # prefix-list
    frr (add|del) routing prefix-list ipv4 <text> description <wildcard>
    frr (add|del) routing prefix-list ipv4 <text> [seq <1-4294967295>] (deny|permit) (<ipv4/prefixlen>|any)
    frr (add|del) routing prefix-list ipv4 <text> [seq <1-4294967295>] (deny|permit) <ipv4/prefixlen> ge <0-32>
    frr (add|del) routing prefix-list ipv4 <text> [seq <1-4294967295>] (deny|permit) <ipv4/prefixlen> ge <0-32> le <1-32>
    frr (add|del) routing prefix-list ipv4 <text> [seq <1-4294967295>] (deny|permit) <ipv4/prefixlen> le <0-32>
    frr (add|del) routing prefix-list ipv4 <text> [seq <1-4294967295>] (deny|permit) <ipv4/prefixlen> le <0-32> ge <1-32>
    frr (add|del) routing prefix-list ipv6 <text> description <wildcard>
    frr (add|del) routing prefix-list ipv6 <text> [seq <1-4294967295>] (deny|permit) (<ipv6/prefixlen>|any)
    frr (add|del) routing prefix-list ipv6 <text> [seq <1-4294967295>] (deny|permit) <ipv6/prefixlen> ge <0-128>
    frr (add|del) routing prefix-list ipv6 <text> [seq <1-4294967295>] (deny|permit) <ipv6/prefixlen> ge <0-128> le <1-128>
    frr (add|del) routing prefix-list ipv6 <text> [seq <1-4294967295>] (deny|permit) <ipv6/prefixlen> le <0-128>
    frr (add|del) routing prefix-list ipv6 <text> [seq <1-4294967295>] (deny|permit) <ipv6/prefixlen> le <0-128> ge <1-128>
    frr del routing prefix-list ipv4 <text> [seq <1-4294967295>]
    frr del routing prefix-list ipv6 <text> [seq <1-4294967295>]
    frr clear ip prefix-list [<prefix-list-v4>|<prefix-list-v4> <ipv4/prefixlen>]
    frr clear ipv6 prefix-list [<prefix-list-v6>|<prefix-list-v6> <ipv6/prefixlen>]
    frr show ip prefix-list [<prefix-list-v4>]
    frr show ip prefix-list <prefix-list-v4> seq <1-4294967295>
    frr show ip prefix-list <prefix-list-v4> <ipv4/prefixlen> [longer|first-match]
    frr show ip prefix-list summary [<prefix-list-v4>]
    frr show ip prefix-list detail [<prefix-list-v4>]
    frr show ipv6 prefix-list [<prefix-list-v6>]
    frr show ipv6 prefix-list <prefix-list-v6> seq <1-4294967295>
    frr show ipv6 prefix-list <prefix-list-v6> <ipv6/prefixlen> [longer|first-match]
    frr show ipv6 prefix-list summary [<prefix-list-v6>]
    frr show ipv6 prefix-list detail [<prefix-list-v6>]

    # route-map
    frr (add|del) routing route-map <text> (deny|permit) <1-65535>
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> call <route-map>
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> continue <0-65535>
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> description <wildcard>
    frr del routing route-map <text> permit <1-65535> on-match (goto <0-65535>|next)
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> match as-path <as-path-access-list>
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> match community <community-list> [exact-match]
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> match extcommunity <extcommunity-list>
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> match interface <interface>
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> match ip (address|next-hop) prefix-len <0-32>
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> match ip (address|next-hop|route-source) prefix-list <prefix-list-v4>
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> match ipv6 address prefix-list <prefix-list-v6>
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> match ipv6 next-hop <ipv6>
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> match local-preference <0-4294967295>
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> match metric <0-4294967295>
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> match origin (egp|igp|incomplete)
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> match peer (<ipv4>|<ipv6>|local)
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> match probability <0-100>
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> match source-protocol (bgp|connected|kernel|ospf|ospf6|static|system)
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> match tag <0-65535>
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> set aggregator as <1-4294967295> <ipv4>
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> set as-path exclude <1-4294967295>
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> set as-path prepend (<1-4294967295>|last-as <1-10>)
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> set atomic-aggregate
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> set comm-list <community-list> delete
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> set community <community> [additive]
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> set community none
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> set extcommunity (rt|soo) <extcommunity>
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> set forwarding-address <ipv6>
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> set label-index <0-1048560>
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> set ip next-hop (<ipv4>|peer-address|unchanged)
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> set ipv6 next-hop (global|local) <ipv6>
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> set ipv6 next-hop peer-address
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> set local-preference <0-4294967295>
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> set metric <0-4294967295>
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> set metric-type (type-1|type-2)
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> set origin (egp|igp|incomplete)
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> set originator-id <ipv4>
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> set src (<ipv4>|<ipv6>)
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> set tag <0-65535>
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> set ipv4 vpn next-hop <ipv4>
    frr (add|del) routing route-map <text> (deny|permit) <1-65535> set weight <0-4294967295>
    frr show route-map [<route-map>]

    # MPLS 'global' configuration
    frr (add|del) mpls lsp <16-1048575> <ip> (<16-1048575>|explicit-null|implicit-null)
    frr (add|del) mpls label bind <ipv4/prefixlen> (<16-1048575>|explicit-null|implicit-null)
    frr (add|del) mpls label global-block <16-1048575> <16-1048575>

    frr show mpls ldp (binding|discovery|interface|neighbor)
    frr show mpls ldp (ipv4|ipv6) (binding|discovery|interface)
    frr show mpls status
    frr show mpls table [<16-1048575>] [json]

    frr (add|del) mpls ldp router-id <ipv4>
    frr (add|del) mpls ldp address-family (ipv4|ipv6) interface <interface>
    frr (add|del) mpls ldp address-family ipv4 discovery transport-address <ipv4>
    frr (add|del) mpls ldp address-family ipv6 discovery transport-address <ipv6>

Options:

    <as-path-access-list>              : An AS_PATH access-list
    <bgppeer>                          : A BGP peer or neighbor
    <community-list>                   : A BGP community-list
    <community>                        : XX:YY where XX and YY are numbers in the range 0-65535
    <extcommunity-list>                : A BGP extended community-list
    <extcommunity>                     : ASN:nn_or_IP-address:nn VPN extended community
    <interface-source>                 : Source interface
    <ipv4-adv-router>                  : OSPF advertising router
    <ipv4-id>                          : OSPFv3 advertising router
    <ipv4-mcast-group>                 : IPv4 multicast group
    <ipv4-source>                      : IPv4 source address
    <peergroup>                        : A BGP peer-group name
    <prefix-list-v4>                   : An IPv4 prefix-list name
    <prefix-list-v6>                   : An IPv6 prefix-list name
    <regex>                            : A regular expression
    <route-map>                        : A route-map name
    <text-password>                    : A password
    <text-peer-group>                  : A BGP peer-group name
    <text-route-map>                   : A route-map name
    Null0                              : Null interface
    abr                                : Area Border Router
    activate                           : Enable the Address Family for this Neighbor
    additive                           : Add to the existing list of communities
    addpath-tx-all-paths               : Use addpath to advertise all paths to a neighbor
    addpath-tx-bestpath-per-AS         : Use addpath to advertise the bestpath per each neighboring AS
    administrative                     : Advertise inability to be a transit router
    adv-router                         : Advertising Router link states
    advertisement-interval             : Minimum interval between sending BGP routing updates
    aggregate-address                  : Configure BGP aggregate entries
    agentx                             : Enable SNMP support for OSPF, OSPFV3, and BGP4 MIBS
    aggregator                         : BGP aggregator attribute
    alerts                             : Immediate action needed
    allowas-in                         : Accept as-path with my AS present in it
    always                             : Always
    always-compare-med                 : Allow comparing MED from different neighbors
    any                                : Any IP prefix
    area                               : An OSPF area
    area-id                            : An OSPF area ID
    as                                 : Autonomous system
    as-external                        : AS-External LSA's
    as-override                        : Override ASNs in outbound updates if aspath equals remote-as
    as-path                            : AS_PATH attribute
    as-path-access-list                : An AS_PATH access-list
    as-set                             : Generate an AS_SET
    asbr                               : Autonomous system border router
    asbr-summary                       : ASBR summary link states
    assert                             : PIM interface assert
    assert-internal                    : PIM interface internal assert state
    assert-metric                      : PIM interface assert metric
    assert-winner-metric               : PIM interface assert winner metric
    atomic-aggregate                   : BGP atomic aggregate attribute
    attribute-unchanged                : BGP attribute is propagated unchanged to this neighbor
    auto-cost                          : Calculate OSPF interface cost according to bandwidth
    autonomous-system                  : AS number
    bestpath                           : BGP bestpath
    bfd                                : Enables BFD support
    bgp                                : Border Gateway Protocol
    blackhole                          : Silently discard pkts when matched
    border-routers                     : OSPF ABR's and ASBR's
    both                               : Both
    broadcast                          : An OSPF broadcast multi-access network
    call                               : Jump to another Route-Map after match+set
    capability                         : Advertise capability to the peer
    client-to-client                   : Configure client to client route reflection
    cluster-id                         : Configure Route-Reflector Cluster-id
    comm-list                          : A BGP community list
    community                          : BGP community attribute
    community-list                     : Add a community list entry
    compare-routerid                   : Compare router-id for identical EBGP paths
    confed                             : Compare path lengths including confederation sets & sequences in selecting a route
    confederation                      : Confederation parameters
    connect                            : BGP connect timer
    connected                          : Connected routes (directly attached subnet or host)
    continue                           : Exit policy on matches
    cost                               : Interface cost
    count                              : Route and packet count data
    critical                           : Critical conditions
    dampening                          : Enable route-flap dampening
    database                           : OSPF database
    dbdesc                             : Database description message
    dd                                 : OSPF Database Description
    dead-interval                      : Interval after which a neighbor is declared dead
    debug                              : Enable debugs
    debugging                          : Debugging messages
    default                            : Configure defaults
    default-originate                  : Originate default route to this neighbor
    deny                               : Deny in routing policy
    description                        : Description
    designated-router                  : PIM interface designated router
    deterministic-med                  : Pick the best-MED path among paths advertised from the neighboring AS
    disable-connected-check            : One-hop away EBGP peer using loopback address
    disable-ebgp-connected-route-check : Disable checking if nexthop is connected on ebgp sessions
    distribute-list                    : Filter updates to/from this neighbor
    dont-capability-negotiate          : Do not perform capability negotiation
    drchoice                           : Designated Router choices
    drpriority                         : PIM Designated Router Election Priority
    dump                               : Dump LSAs
    ebgp-multihop                      : Allow EBGP neighbors not on directly connected networks
    egp                                : Exterior Gateway Protocol
    emergencies                        : System is unusable
    enforce-first-as                   : Enforce the first AS for EBGP routes
    enforce-multihop                   : Enforce EBGP neighbors perform multihop
    errors                             : Error conditions
    event                              : Event information
    events                             : Event information
    exact-match                        : Do exact matching of communities
    examine                            : Examine
    exclude                            : Exclude from the as-path
    expanded                           : An expanded community-list
    extcommunity                       : An extended community
    extcommunity-list                  : An extended community list
    extended                           : Send Extended Community attributes
    extended-nexthop                   : Advertise extended next-hop capability to the peer
    external                           : An external BGP peer, external OSPF LSA, etc
    external-1                         : OSPF Type-1 External routes
    external-2                         : OSPF Type-2 External routes
    fast-external-failover             : Immediately reset session if a link to a directly connected external peer goes down
    file                               : Logging to file
    filter-list                        : Apply an as-path access-list
    first-match                        : First matched prefix
    flooding                           : LSA Flooding
    force                              : Set the next hop to self for reflected routes
    forwarding-address                 : Forwarding Address
    ge                                 : Greater than or equal to
    generate                           : LSA Generation
    global                             : IPv6 global address
    goto                               : Goto Clause number
    graceful-restart                   : Graceful restart capability parameters
    group-type                         : PIM group type
    groups                             : IGMP groups information
    hello                              : PIM interface hello information
    hello-interval                     : Time between HELLO packets
    hops                               : TTL hop count
    ibgp                               : An internal BGP peer
    ifmtu                              : Interface MTU
    igmp                               : Internet Group Management Protocol
    ignore                             : Ignore as-path length in selecting a route
    igp                                : Interior Gateway Protocol
    import-table                       : Import routes from non-main kernel table
    in                                 : Apply inbound
    inbound                            : Apply inbound
    incomplete                         : Unknown origin
    informational                      : Informational messages
    install                            : LSA Install/Delete
    instance-id                        : Instance identifier
    inter-area                         : OSPF Inter-Area routes
    inter-prefix                       : Inter-Prefix LSA's
    inter-router                       : Inter-Router LSA's
    interface                          : An interface, such as swp1, swp2, etc.
    internal                           : An internal BGP peer
    internet                           : BGP community, matches all routes
    intra-area                         : OSPF Intra-Area routes
    intra-prefix                       : Intra-Prefix LSA's
    ip                                 : Internet Protocol version 4
    ipv4                               : Internet Protocol version 4
    ipv4-unicast                       : IPv4 Unicast address family
    labeled-unicast                    : Address Family modifier
    label-index                        : Label index value
    ipv6                               : Internet Protocol version 6
    ism                                : Interface State Machine
    join                               : PIM interface join information
    jp-override-interval               : PIM interface J/P override interval
    keepalives                         : BGP keepalives
    kernel                             : Kernel routes (not installed via the zebra RIB)
    lan-prune-delay                    : PIM neighbors LAN prune delay parameters
    last-as                            : Use the peer's AS-number
    le                                 : Less than or equal to
    limit                              : Maximum number of BGP Dynamic Neighbors that can be created
    linkstate                          : Display linkstate routing table
    linkstate-id                       : Search by Link state ID
    listen                             : Configure BGP Dynamic Neighbors
    local                              : IPv6 local address
    local-AS                           : BGP community, restrict routes to our confederations sub-AS
    local-as                           : Specify a local-as number
    local-membership                   : PIM interface local-membership
    local-preference                   : BGP local-preference attribute
    log                                : Logging control
    log-adjacency-changes              : Log neighbor up/down and reset reason
    log-neighbor-changes               : Log neighbor up/down and reset reason
    longer                             : Show route matching the specified Network/Mask pair only
    longer-prefixes                    : Show route matching the specified Network/Mask pair only
    ls-ack                             : Link State Acknowledgement
    ls-request                         : Link State Request
    ls-update                          : Link State Update
    lsa                                : Link State Advertisement
    lsack                              : Link State Acknowledgement
    lsreq                              : Link State Request
    lsupdate                           : Link State Update
    match                              : Match values from routing table
    max-age                            : LSAs in MaxAge list
    max-med                            : Advertise routes with max-med
    maximum-paths                      : Forward packets over multiple paths
    maximum-prefix                     : Maximum number of prefix accept from this peer
    med                                : MED attribute
    memory                             : Memory statistics
    message                            : OSPFv3 message
    metric                             : Metric value for destination routing protocol
    metric-type                        : Type of metric
    min-arrival                        : Minimum delay in receiving new version of a LSA
    missing-as-worst                   : Treat missing MED as the least preferred one
    mroute                             : Static unicast routes in MRIB for multicast RPF lookup
    msdp                               : Multicast Source Discovery Protocol
    msec                               : Millisecond
    mtu-ignore                         : Ignore MTU
    multicast                          : Multicast
    traffic                            : Protocol Packet counters
    multihop                           : Neighbor is not directly connected
    multipath                          : BGP paths flagged as multipath
    multipath-relax                    : Allow load sharing across routes that have different AS paths (but same length)
    nd                                 : Neighbor Discovery
    neighbor                           : A BGP, OSPF, PIM, etc neighbor
    neighbor-events                    : Neighbor state transition events
    network                            : An ipv4 or ipv6 network
    next                               : Next route-map clause
    next-hop                           : Next-hop address of route
    next-hop-self                      : Disable the next hop calculation for this neighbor
    nht                                : Next Hop Tracking
    no-advertise                       : BGP community, do not advertise to anyone
    no-as-set                          : Do not generate an AS_SET
    no-export                          : BGP community, do not advertise to eBGP peer
    no-summary                         : Do not inject inter-area routes into stub
    non-broadcast                      : An OSPF NBMA network
    none                               : No community attribute
    notifications                      : Normal but significant conditions
    nsm                                : Neighbor State Machine
    nssa                               : OSPF nssa information
    nssa-external                      : NSSA external link state
    on-match                           : Exit policy on matches
    opaque-area                        : Link area Opaque-LSA
    opaque-as                          : Link AS Opaque-LSA
    opaque-link                        : Link local Opaque-LSA
    origin                             : BGP origin attribute or the originating AS in the AS_PATH
    originate                          : Originate
    originator-id                      : BGP ORIGINATOR_ID attribute
    ospf                               : Open Shortest Path First (OSPFv2)
    ospf6                              : Open Shortest Path First (OSPFv3)
    out                                : Apply outbound
    override-capability                : Override capability negotiation result
    packet                             : Packet
    packets                            : Packets
    parameters                         : IGMP parameters information
    passive                            : Do not send open messages to this neighbor
    passive-interface                  : Suppress routing updates on an interface
    password                           : Set a password
    peer                               : peer address
    peer-address                       : Use peer address (for BGP only)
    peer-group                         : Member of the peer-group
    permit                             : Permit in routing policy
    pim                                : Protocol Independent Multicast
    point-to-multipoint                : An OSPF point-to-multipoint network
    point-to-point                     : An OSPF point-to-point network
    port                               : Port
    precision                          : Set the timestamp precision
    prefix                             : An IPv4 or IPv6 prefix
    prefix-len                         : Match prefix length of ip address
    prefix-list                        : Filter updates to/from this neighbor
    prepend                            : Prepend to the as-path
    priority                           : priority
    probability                        : Match portion of routes defined by percentage value
    process                            : Debug Detailed SPF Process
    protocol                           : Filter routing info exchanged between zebra and protocol
    ptm-enable                         : Enable neighbor check with specified topology
    query-interval                     : IGMP host query interval
    query-max-response-time            : IGMP max query response value (seconds)
    ra-interval                        : Router Advertisement interval
    ra-lifetime                        : Router lifetime
    range                              : Add a listening range for Dynamic Neighbors
    recv                               : Packet received
    redistribute                       : Redistribute information from another routing protocol
    reference-bandwidth                : Use reference bandwidth method to assign OSPF cost
    refresh                            : LSA Refresh
    reject                             : Emit an ICMP unreachable when matched
    remote-as                          : Specify a BGP neighbor
    remove-private-AS                  : Remove private ASNs in outbound updates
    replace-AS                         : Replace private ASNs with our ASN in outbound updates
    retransmit-interval                : Time between retransmitting lost link state advertisements
    rip                                : Routing Information Protocol (RIP)
    ripng                              : Routing Information Protocol Next Generation
    route                              : Static routes
    route-map                          : Route-map
    route-reflector                    : Allow modifications made by out route-map
    route-reflector-client             : Configure a neighbor as Route Reflector client
    route-server-client                : Configure a neighbor as Route Server client
    route-source                       : Match advertising source address of route
    router                             : L3 routing protocols or OSPF LSA type
    router-id                          : Router identifier
    rp                                 : Rendevous Point
    rp-info                            : PIM RP information
    rpf                                : PIM cached source rpf information
    rt                                 : Route Target extended community
    secondary                          : The second or next highest in rank or importance
    self-originate                     : Self-originated link states
    self-originated                    : Self-originated link states
    send                               : Packet sent
    send-community                     : Send Community attribute to this neighbor
    seq                                : Sequence number
    set                                : Set values in destination routing protocol
    shutdown                           : Administratively shut down this neighbor
    soft                               : Soft reconfig inbound and outbound updates
    soft-reconfiguration               : Per neighbor soft reconfiguration
    solo                               : Solo peer - part of its own update group
    soo                                : Site-of-Origin extended community
    source-protocol                    : Match protocol via which the route was learnt
    sources                            : IGMP sources information
    spf                                : OSPF SPF timers
    src                                : Source
    ssm                                : Source Specific Multicast
    standard                           : Standard BGP communities
    state                              : State information
    static                             : Statically configured routes
    strict-capability-match            : Strict capability negotiation match
    stub                               : Configure OSPF area as stub
    stub-router                        : Make router a stub router
    summary                            : Summary
    summary-only                       : Filter more specific routes from updates
    supernets-only                     : Show only supernets
    suppress-ra                        : Suppress IPv6 Router Advertisement
    table                              : Non-main Kernel Routing Table
    table-map                          : BGP table to RIB route download filter
    tag                                : Set tag for this route
    throttle                           : Throttling adaptive timer
    time                               : Time
    timers                             : Routing protocol timers
    timestamp                          : Timestamp configuration
    trace                              : IGMP/PIM internal daemon activity
    transmit-delay                     : Transmit delay of this interface
    tree                               : Show SPF tree
    ttl-security                       : Specify the maximum number of hops to the BGP peer
    type-1                             : OSPF6 external type 1 metric
    type-2                             : OSPF6 external type 2 metric
    unchanged                          : Don't modify existing Next hop address
    unicast                            : Unicast Address Family Modifier
    unknown                            : Unknown LSA's
    unsuppress-map                     : Route-map to selectively unsuppress suppressed routes
    update-delay                       : Force initial delay for best-path and updates
    update-groups                      : BGP update groups
    update-source                      : Source of routing updates
    updates                            : BGP updates
    upstream                           : An upstream interface
    upstream-join-desired              : PIM upstream join-desired
    upstream-rpf                       : PIM upstream source rpf
    v6only                             : Enable BGP with v6 link-local only
    vpnv4                              : Vpnv4 address family
    vrf                                : Virtual Routing and Forwarding
    warnings                           : Warning conditions
    weight                             : Set default weight for routes from this neighbor
    zebra                              : Zebra information
    access-list                        : An access list
    Null0                              : Null interface
    routing                            : Routing protocols
    sa                                 : Source address
    sm                                 : Pim sparse mode
    2                                  : IGMP version 2, OSPF metric-type 2, etc
    3                                  : IGMP version 3
    join-prune-interval                : Join Prune Send Interval
    keep-alive-timer                   : Keep alive Timer
    register-suppress-time             : Register Suppress Timer
    ecmp                               : Equal Cost Multi-Path
    rebalance                          : Equal Cost Multi-Path Rebalance
    mesh-group                         : MSDP mesh-group
    member                             : mesh group member
    rd                                 : Route Distinguisher
    route-target                       : Route Target
    <rd>                               : A.B.C.D:2-byte-number, 4-byte-number:2-byte-number or 2-byte-number:4-byte-number
    <route-target>                     : 4-byte-number:2-byte-number or 2-byte-number:4-byte-number
    evpn                               : Ethernet VPN
    l2vpn                              : Layer-2 VPN
    import-rt                          : Import Route Target
    export-rt                          : Export Route Target
    vni                                : VXLAN Network Identifier
    type                               : Type
    macip                              : MAC-IP (Type-2) route
    mac                                : Media Access Control
    vtep                               : VXLAN Tunnel End Point
    advertise-all-vni                  : Advertise All local VNIs
    advertise-default-gw               : Advertise All defualt g/w mac-ip routes in EVPN
    import                             : Import
    export                             : Export
    use-source                         : Specify source
    authentication                     : Enable authentication
    message-digest                     : Use message-digest authentication
    message-digest-key                 : Message digest authentication password (key)
    md5                                : MD5 algorithm
    joins                              : PIM Join/Prune protocol packets
    register                           : PIM Register/Reg-Stop protocol packets
    mroute                             : A static unicast route into MRIB for multicast RPF lookup
    debugs                             : Debugs
    syslog                             : Set syslog logging
    line                               : A terminal line
    vty                                : Virtual terminal
    exec-timeout                       : Set the EXEC timeout
    service                            : Service
    integrated-vtysh-config            : Write configuration into integrated file
    import-table                       : Import routes from non-main kernel table
    fpm                                : Debug zebra FPM events
    mpls                               : Multiprotocol Label Switching
    rib                                : Routing Information Base
    detailed                           : Detailed
    table                              : Table
    explicit-null                      : Use Explicit-Null label
    implicit-null                      : Use Implicit-Null label
    label                              : MPLS label
    ldp                                : Label Distribution Protocol
    binding                            : Binding
    discovery                          : Discovery
    address-family                     : address-family
    transport-address                  : Specify transport address for TCP connection
    bind                               : Establish FEC to label binding
    global-block                       : Configure global label block
    lsp                                : Label Switched Path
    nexthop-lookup                     : PIM cached nexthop rpf lookup
    infinite                           : Infinite valid lifetime
    no-autoconfig                      : Do not use prefix for autoconfiguration
    off-link                           : Do not use prefix for onlink determination
    router-address                     : Set Router Address flag
    advertised-routes                  : Display routes advertised to a BGP neighbor
    received-routes                    : Display routes learned from neighbor, including denied routes
    routes                             : Display routes learned from neighbor

"""

# Syntax for access-list...we decided to leave these out since prefix-lists
# are a much better way for filtering routes.
'''
    # access-list
    frr (add|del) access-list (<100-199>|<2000-2699>) (deny|permit) ip <ipv4-source> <ipv4-source-wildcard> <ipv4-dest> <ipv4-dest-wildcard>
    frr (add|del) access-list (<100-199>|<2000-2699>) (deny|permit) ip <ipv4-source> <ipv4-source-wildcard> any
    frr (add|del) access-list (<100-199>|<2000-2699>) (deny|permit) ip <ipv4-source> <ipv4-source-wildcard> host <ipv4-dest>
    frr (add|del) access-list (<100-199>|<2000-2699>) (deny|permit) ip any <ipv4-dest> <ipv4-dest-wildcard>
    frr (add|del) access-list (<100-199>|<2000-2699>) (deny|permit) ip any any
    frr (add|del) access-list (<100-199>|<2000-2699>) (deny|permit) ip any host <ipv4-dest>
    frr (add|del) access-list (<100-199>|<2000-2699>) (deny|permit) ip host <ipv4-source> <ipv4-dest> <ipv4-dest-wildcard>
    frr (add|del) access-list (<100-199>|<2000-2699>) (deny|permit) ip host <ipv4-source> any
    frr (add|del) access-list (<100-199>|<2000-2699>) (deny|permit) ip host <ipv4-source> host <ipv4-dest>
    frr (add|del) access-list (<1-99>|<1300-1999>) (deny|permit) <ipv4-source> [<ipv4-source-wildcard>]
    frr (add|del) access-list (<1-99>|<1300-1999>) (deny|permit) (any|host <ipv4-source>)
    frr (add|del) access-list <text> (deny|permit) (any|<ipv4/prefixlen>|<ipv4/prefixlen> exact-match)
    frr (add|del) ipv6 access-list <text> (deny|permit) (any|<ipv6/prefixlen>|<ipv6/prefixlen> exact-match)
    frr show ip access-list [<1-2699>|<text>]
    frr show ipv6 access-list [<text>]
'''

from collections import OrderedDict
from copy import deepcopy
from ipaddr import IPNetwork
from network_docopt import NetworkDocopt, get_network_docopt_info, sort_for_humans, Word, Node
from nclu import (print_diff,
    files_match,
    chown_chgroup,
    file_contains,
    ifname_is_valid,
    ifname_is_globish,
    ifname_is_glob,
    ifname_expand_glob,
    PENDING_COMMIT_NOCHANGE,
    PENDING_COMMIT_CHANGE,
    PENDING_COMMIT_ERROR,
    globify_interface_commands,
    glob_to_numbers,
    Capturing,
    print_pending_add_del_commands,
    service_is_active)
from os import listdir
from re import search as re_search
from pprint import pformat, pprint
from shutil import copyfile
from subprocess import check_output, STDOUT, CalledProcessError
import json
import logging
import logging.handlers
import os
import re
import sys


sys.path.append('/usr/lib/frr/')
frr_reload = __import__("frr-reload")  # use __import__ for files with a hyphen in the name
logging.getLogger('frr-reload').setLevel(logging.WARNING)

log = logging.getLogger('netd')

if os.path.exists('/var/run/frr/.qcompat'):
    etc_frr = '/etc/quagga/'
    etc_frr_daemons = '/etc/quagga/daemons'
    etc_frr_frr_conf = '/etc/quagga/Quagga.conf'
    etc_frr_frr_conf_backup = '/etc/quagga/Quagga.conf.backup'
    private_pending_conf = '/var/run/nclu/quagga/pending.conf'
    service_name = 'quagga'
else:
    etc_frr = '/etc/frr/'
    etc_frr_daemons = '/etc/frr/daemons'
    etc_frr_frr_conf = '/etc/frr/frr.conf'
    etc_frr_frr_conf_backup = '/etc/frr/frr.conf.backup'
    private_pending_conf = '/var/run/nclu/frr/pending.conf'
    service_name = 'frr'

usr_lib_frr_pimd = '/usr/lib/frr/pimd'
frr_reload_py = '/usr/lib/frr/frr-reload.py'
public_pending_conf = '/var/run/nclu/frr_pending.conf'

def build_doc_string():
    pass


def sanity_check_ip_host_vs_subnet(args, cli):
    """
    Certain commands like "ip route" will convert a host IP to the subnet.
    Look for this so we can give the user a heads up that this is happening.
    """
    ipv4_str = args.get('<ipv4/prefixlen>')

    if ipv4_str:
        ipv4 = IPNetwork(ipv4_str)
        subnet = "%s/%s" % (ipv4.network, ipv4.prefixlen)

        if ipv4_str != subnet:
            args['<ipv4/prefixlen>'] = subnet
            print "NOTE: This command expects a subnet not a host IP so %s was converted to %s" % (ipv4_str, subnet)

            for (index, node) in enumerate(cli.match):
                if '<ipv4/prefixlen>' in str(node):
                    node.value.value = subnet
                    node.value.value_raw = subnet

    ipv6_str = args.get('<ipv6/prefixlen>')

    if ipv6_str:
        ipv6 = IPNetwork(ipv6_str)
        subnet = "%s/%s" % (ipv6.network, ipv6.prefixlen)

        if ipv6_str != subnet:
            args['<ipv6/prefixlen>'] = subnet
            print "NOTE: This command expects a subnet not a host IP so %s was converted to %s" % (ipv6_str, subnet)

            for (index, node) in enumerate(cli.match):
                if '<ipv6/prefixlen>' in str(node):
                    node.value.value = subnet
                    node.value.value_raw = subnet


def delete_this_line(line, del_line, del_line_vni, vni, exact_match):
    """
    Return True if this line should be deleted, else return False
    """

    if del_line_vni:
        if del_line_vni == vni:
            # If the user is deleting the entire vni under "address-family l2vpn evpn"
            # then del_line will be None
            if del_line is None:
                return True

            if exact_match and line.strip() == del_line:
                return True

            if not exact_match and line.strip().startswith(del_line):
                return True

    elif del_line_vni is None and vni:
        return False

    elif del_line is None:
        return False

    else:
        if exact_match and line.strip() == del_line:
            return True

        if not exact_match and line.strip().startswith(del_line):
            return True

        # If you remove "ip igmp" from an interface frr will remove all of
        # the other "ip igmp XYZ" commands under that interface so go ahead
        # and remove those commands from the pending buffer.
        if del_line == 'ip igmp' and line.lstrip().startswith('ip igmp'):
            return True

        # prefix-lists are special
        if ((del_line.startswith('ip prefix-list') and line.startswith('ip prefix-list') or
            (del_line.startswith('ipv6 prefix-list') and line.startswith('ipv6 prefix-list')))):

            re_name_only = re.search('(ip|ipv6) prefix-list \w+$', del_line)
            re_name_seq_only = re.search('(ip|ipv6) prefix-list \w+ seq \d+$', del_line)

            # delete the entire prefix-list
            # delete an entry in the prefix-list by seq number
            if re_name_only or re_name_seq_only:
                if line.startswith(del_line + ' '):
                    return True

            # delete an entry in the prefix-list by value
            elif ' seq ' not in del_line:
                re_seq = re.search('^(.*) seq \d+ (.*)$', line)

                if re_seq:
                    line_minus_seq = re_seq.group(1) + ' ' + re_seq.group(2)

                    if del_line == line_minus_seq:
                        return True

    return False


def frr_reload_exception_to_string(filename, e):
    if hasattr(e, 'output'):
        return "Failed to load %s due to %s\n\n%s output\n=======================\n%s\n" %\
            (filename, str(e), frr_reload_py, e.output)
    else:
        return "Failed to load %s due to %s" % (filename, str(e))


class ConfigWrapper(object):

    def __init__(self, show_linux_command, color_diffs):
        self.init_working_dir()
        self.show_linux_command = show_linux_command
        self.color_diffs = color_diffs

    def init_working_dir(self):
        self.working_dir = '/var/run/nclu/%s/' % service_name
        self.baseline_conf = os.path.join(self.working_dir, 'baseline.conf')
        self.pending_conf = os.path.join(self.working_dir, 'pending.conf')
        self.pending_baseline_conf = os.path.join(self.working_dir, 'pending_baseline.conf')
        self.baseline_daemons = os.path.join(self.working_dir, 'baseline.daemons')
        self.pending_daemons = os.path.join(self.working_dir, 'pending.daemons')

        if not os.path.isdir(self.working_dir):
            os.makedirs(self.working_dir, mode=0600)

    def get_managed_files(self):
        files = []
        files.append(etc_frr_daemons)
        files.append(etc_frr_frr_conf)
        return files

    def get_service_cmd_for_file(self, filename):
        cmds = []

        if filename == etc_frr_daemons:
            cmds.append(('/bin/systemctl', 'enable', service_name))
            cmds.append(('/bin/systemctl', 'reset-failed', service_name))
            cmds.append(('/bin/systemctl', 'restart', service_name))

        elif filename == etc_frr_frr_conf:
            cmds.append(('/bin/systemctl', 'enable', service_name))
            cmds.append(('/bin/systemctl', 'reset-failed', service_name))
            cmds.append(('/bin/systemctl', 'reload', service_name))

        return cmds

    def get_daemons(self):
        daemons = {}

        # Create an ordered dictionary from the uncommented "<daemon>=<yes|no>" lines in the file.
        if os.path.exists(self.pending_daemons):
            with open(self.pending_daemons, 'r') as f:
                pending_daemons_text = f.read()
            daemon_settings = re.findall(r"""^\s*(\w+)\s*=\s*(yes|no)\s*$""", pending_daemons_text, re.I | re.M)
            daemons = OrderedDict(daemon_settings)

        return daemons

    def init_guts(self):
        """
        Look for problems with frr's configuration files, and correct them if possible.
        Return True if there are no problems (possibly after correction); otherwise, return
        False.
        """
        global etc_frr, etc_frr_daemons, etc_frr_frr_conf, etc_frr_frr_conf_backup, private_pending_conf, service_name

        # If quagga was running when netd started but frr is running now then we
        # need to switch gears and work against /etc/frr/ instead of /etc/quagga/
        if os.path.exists('/var/run/frr/.qcompat'):
            if service_name == 'frr':
                log.info("changing from frr to quagga")
                etc_frr = '/etc/quagga/'
                etc_frr_daemons = '/etc/quagga/daemons'
                etc_frr_frr_conf = '/etc/quagga/Quagga.conf'
                etc_frr_frr_conf_backup = '/etc/quagga/Quagga.conf.backup'
                private_pending_conf = '/var/run/nclu/quagga/pending.conf'
                service_name = 'quagga'
                self.init_working_dir()
        else:
            if service_name == 'quagga':
                log.info("changing from quagga to frr")
                etc_frr = '/etc/frr/'
                etc_frr_daemons = '/etc/frr/daemons'
                etc_frr_frr_conf = '/etc/frr/frr.conf'
                etc_frr_frr_conf_backup = '/etc/frr/frr.conf.backup'
                private_pending_conf = '/var/run/nclu/frr/pending.conf'
                service_name = 'frr'
                self.init_working_dir()

        # If the /etc/frr/ directory isn't even there then return False
        if not os.path.isdir(etc_frr):
            log.info("frr wrapper could not initialize, %s is missing" % etc_frr)
            return False

        if not os.path.isdir(self.working_dir):
            os.makedirs(self.working_dir, mode=0600)

        # If /etc/frr/daemons does not exist or is empty, create a default file.
        if not os.path.exists(etc_frr_daemons) or os.path.getsize(etc_frr_daemons) == 0:
            with open(etc_frr_daemons, 'w') as f:
                f.write('zebra=yes\n')

        # If /etc/frr/frr.conf does not exist, create a default file.
        if not os.path.isfile(etc_frr_frr_conf):
            with open(etc_frr_frr_conf, 'w') as f:
                f.write('service integrated-vtysh-config\n')

        if not os.path.isfile(self.baseline_daemons):
            copyfile(etc_frr_daemons, self.baseline_daemons)
            copyfile(self.baseline_daemons, self.pending_daemons)

        # FRR is not running so we must start it.
        if not service_is_active(service_name):
            daemons = self.get_daemons()

            # zebra must always be enabled.
            if daemons.get('zebra', '').lower() != 'yes':
                daemons['zebra'] = 'yes'

                # Overwrite the old "daemons" file.
                with open(etc_frr_daemons, 'w') as f:
                    for (daemon, daemon_value) in daemons.iteritems():
                        f.write("{0}={1}\n".format(daemon, daemon_value))
                copyfile(etc_frr_daemons, self.baseline_daemons)
                copyfile(self.baseline_daemons, self.pending_daemons)

            for cmd in (('/bin/systemctl', 'enable', service_name),
                        ('/bin/systemctl', 'reset-failed', service_name),
                        ('/bin/systemctl', 'start', service_name)):

                try:
                    check_output(cmd, stderr=STDOUT)
                except CalledProcessError as e:
                    print "ERROR running %s\n" %  ' '.join(cmd)
                    print e.output
                    return False

        # In the beginning, our staging file is a "marked" copy of frr.conf.
        # "marked" means that each context ends with the word "end".
        if not os.path.isfile(self.pending_conf) or not os.path.isfile(self.pending_baseline_conf):

            # If this fails it means frr isn't running (it should be)
            try:
                output = check_output(['/usr/bin/vtysh', '-c', 'wr mem'],
                                       stderr=STDOUT).splitlines()
            except CalledProcessError as e:
                print "ERROR running 'wr mem'\n"
                print e.output
                return False

            try:
                output = check_output(['/usr/bin/vtysh',
                                       '-m',
                                       '-f',
                                       etc_frr_frr_conf],
                                       stderr=STDOUT).splitlines()
            except CalledProcessError as e:
                print "ERROR marking %s\n" % etc_frr_frr_conf
                print e.output
                return False
            else:
                with open(self.pending_conf, 'w') as fh:
                    for line in output:
                        if line != '!':
                            fh.write(line + '\n')

                with open(self.pending_baseline_conf, 'w') as fh:
                    for line in output:
                        if line != '!':
                            fh.write(line + '\n')

        # Use baseline.conf to determine later if frr.conf has been modified.
        if not os.path.isfile(self.baseline_conf):
            copyfile(etc_frr_frr_conf, self.baseline_conf)

        # Verify that frr's configuration files have not been modified.
        if not files_match(etc_frr_daemons, self.baseline_daemons):
            print "\nERROR: %s was modified by another user" % etc_frr_daemons
            print_pending_add_del_commands()
            return False

        if not files_match(etc_frr_frr_conf, self.baseline_conf):
            print "\nERROR: %s was modified by another user" % etc_frr_frr_conf
            print_pending_add_del_commands()
            return False

        return True

    def look_for_line(self, target_line, exact_match):
        """
        Look for the target_line in pending_baseline_conf and pending_conf
        If we find the line, return it else return None
        """
        lines = []

        if os.path.exists(self.pending_baseline_conf):
            with open(self.pending_baseline_conf, 'r') as fh:
                for line in fh.readlines():
                    line = line.strip()

                    if exact_match:
                        if line == target_line:
                            lines.append(line)
                    else:
                        if line.startswith(target_line):
                            lines.append(line)

        if os.path.exists(self.pending_conf):
            with open(self.pending_conf, 'r') as fh:
                for line in fh.readlines():
                    line = line.strip()

                    if exact_match:
                        if line == target_line:
                            lines.append(line)
                    else:
                        if line.startswith(target_line):
                            lines.append(line)

        return lines

    def get_bgp_asn(self, vrf):
        asn = None

        if vrf:
            vrf = "vrf %s" % vrf

        if os.path.exists(self.pending_conf):
            with open(self.pending_conf, 'r') as fh:
                for line in fh.readlines():
                    if line.startswith('router bgp'):
                        line = line.strip()
                        if (vrf and line.endswith(vrf)) or (not vrf and 'vrf' not in line):
                            asn = int(line.split()[2])
                            break

        return asn

    def neighbor_interface(self, neighbor):
        """
        Return True if 'neighbor' is already configured as an interface neighbor
        """
        target_line = 'neighbor %s interface' % neighbor

        if self.look_for_line(target_line, False):
            return True
        else:
            return False

    def neighbor_remote_as(self, neighbor):
        """
        Return True if 'neighbor' already has a remote-as configured
        """

        # First see if the neighbor has remote-as configured directly on the neighbor
        neighbor_remote_as = 'neighbor %s remote-as' % neighbor

        if self.look_for_line(neighbor_remote_as, False):
            return True

        # If not, see if the peer is a member of a peer-group where the
        # peer-group has remote-as configured
        neighbor_peer_group = 'neighbor %s peer-group ' % neighbor
        neighbor_peer_group = self.look_for_line(neighbor_peer_group, False)

        if neighbor_peer_group:
            neighbor_peer_group = neighbor_peer_group[0]

            if neighbor_peer_group.endswith('peer-group'):
                peer_group = neighbor_peer_group.split()[1]
            else:
                peer_group = neighbor_peer_group.split()[3]
            peer_group_remote_as = 'neighbor %s remote-as' % peer_group

            if self.look_for_line(peer_group_remote_as, False):
                return True

        return False

    def del_all(self, cli):

        if self.init_guts():
            with open(self.pending_daemons, 'w') as f:
                f.write("zebra=yes\n")

            with open(self.pending_conf, 'w') as f:
                f.write("service integrated-vtysh-config\n")

    def edit_pending_file(self, target_context, target_subcontext, add_line, del_line, exact_match, give_feedback=True):
        """
        target_context - A string; the first subcommand of "net" (i.e. add, del, etc.)

        target_subcontext - A string; the minor subcommand of target_context (i.e. bgp in "net del bgp ...")

        add_line and del_line are None or a string.  Only one should not be None.  The string is the
            user's command.
        """
        assert (add_line is None) != (del_line is None)  # XOR

        if not self.init_guts():
            return False  # Something is wrong with one or both of frr's configuration files.

        if target_subcontext:
            subcontext_display = "\n %s\n" % target_subcontext
        else:
            subcontext_display = ''                      # Not all contexts have a subcontext.

        if add_line is not None:
            action = 'Adding'
            add_line = add_line.rstrip()
            line_display = "\n %s\n" % add_line
            found_del_line = True

            if (target_context and
                target_context.startswith('router bgp') and
                target_subcontext == 'address-family l2vpn evpn'):

                # evpn/vni special exception...at the time evpn was coded it was
                # the first frr feature wrappered by nclu to use a
                #
                #   context -> sub-context -> sub-sub-context
                #
                # model where "vni X" is the sub-sub-context within "address-family evpn".
                # As a quick solution for supporting this we always specify the vni line.
                re_vni = re.search(r'(vni \d+) (.*)', add_line)

                if re_vni:
                    add_line = "%s\n    %s" % (re_vni.group(1), re_vni.group(2))

        del_line_vni = None

        if del_line is not None:
            action = 'Deleting'
            del_line = del_line.strip()
            line_display = "\n %s\n" % del_line
            found_del_line = False

            # Example: "vni 10 rd 1:1"
            if del_line.startswith('vni '):
                del_line_list = del_line.split()
                del_line_vni = del_line_list[1]
                # If the user is deleting the entire vni subsubcontext then make del_line None
                del_line = ' '.join(del_line_list[2:]) if len(del_line_list) > 2 else None

        log.debug("""
action     : {0}
context    : {1}
subcontext : {2}
line       : {3}\n""".format(action, target_context, subcontext_display, line_display))

        #############################################################################################
        # Stage an update to the daemon's file.  Most of the time, we won't need to modify this file.
        #############################################################################################

        rewrite_daemons = False  # Do we need to modify the "daemons" file?
        daemons = self.get_daemons()

        # zebra must always be enabled.
        if daemons.get('zebra', '').lower() != 'yes':
            daemons['zebra'] = 'yes'
            rewrite_daemons = True

        if add_line:

            # If this is a multicast related command, pimd must be enabled.
            if ('igmp ' in add_line or ' igmp' in add_line or
                'pim ' in add_line or ' pim' in add_line or
                'multicast' in add_line or 'msdp' in add_line) and os.path.exists(usr_lib_frr_pimd):
                if daemons.get('pimd', '').lower() != 'yes':
                    daemons['pimd'] = 'yes'
                    rewrite_daemons = True

            if target_context:

                # Enable bgpd, if needed.
                if target_context.startswith('router bgp'):
                    if daemons.get('bgpd', '').lower() != 'yes':
                        daemons['bgpd'] = 'yes'
                        rewrite_daemons = True

                # Enable ospfd, if needed.
                elif target_context == 'router ospf':
                    if daemons.get('ospfd', '').lower() != 'yes':
                        daemons['ospfd'] = 'yes'
                        rewrite_daemons = True

                elif target_context == 'router ospf6':
                    if daemons.get('ospf6d', '').lower() != 'yes':
                        daemons['ospf6d'] = 'yes'
                        rewrite_daemons = True

                elif target_context.startswith('interface'):
                    if 'ospf6' in add_line:
                        if daemons.get('ospf6d', '').lower() != 'yes':
                            daemons['ospf6d'] = 'yes'
                            rewrite_daemons = True

                    elif 'ospf' in add_line:
                        if daemons.get('ospfd', '').lower() != 'yes':
                            daemons['ospfd'] = 'yes'
                            rewrite_daemons = True

                    elif ' pim ' in add_line or ' igmp ' in add_line:
                        if daemons.get('pimd', '').lower() != 'yes':
                            daemons['pimd'] = 'yes'
                            rewrite_daemons = True

                elif target_context.startswith('mpls ldp'):
                    if daemons.get('ldpd', '').lower() != 'yes':
                        daemons['ldpd'] = 'yes'
                        rewrite_daemons = True

            else:
                # target_context is None for commands that go in the global
                # context (static routes, debugs, etc.).
                if 'debug bgp ' in add_line:
                    if daemons.get('bgpd', '').lower() != 'yes':
                        daemons['bgpd'] = 'yes'
                        rewrite_daemons = True

                elif 'debug ospf ' in add_line:
                    if daemons.get('ospfd', '').lower() != 'yes':
                        daemons['ospfd'] = 'yes'
                        rewrite_daemons = True

                elif 'debug ospf6 ' in add_line:
                    if daemons.get('ospf6d', '').lower() != 'yes':
                        daemons['ospf6d'] = 'yes'
                        rewrite_daemons = True

                elif 'debug pim ' in add_line or 'debug mroute' in add_line:
                    if daemons.get('pimd', '').lower() != 'yes':
                        daemons['pimd'] = 'yes'
                        rewrite_daemons = True

        if rewrite_daemons:
            # Overwrite the old "daemons" file.
            with open(self.pending_daemons, 'w') as f:
                for (daemon, daemon_value) in daemons.iteritems():
                    f.write("{0}={1}\n".format(daemon, daemon_value))

        #############################################################################################
        # Stage configuration changes to pending_conf.
        #############################################################################################
        use_exit_address_family = bool(target_subcontext and target_context.startswith('router bgp'))

        # r+ opens the file for reading and writing.
        with open(self.pending_conf, 'r+') as f:

            # Read all lines, and reset the file pointer back to
            # the start of the file.
            config_lines = list(f)
            f.seek(0)

            vni = None
            state = 'LIMBO'
            subcontext_lines = []

            # Write all of the lines back to the file, but add the add_line
            # (if there is one) at the end of its context/subcontext.
            for line in config_lines:
                line = line.rstrip()

                if line.lstrip().startswith('vni '):
                    vni = line.split()[1]

                if state == 'LIMBO':
                    if target_context:
                        if line == target_context:
                            state = 'IN_CONTEXT'
                    # If we are adding lines at the global level (static routes, debugs, etc),
                    # target_context will be None.  Do nothing for this case.

                elif state == 'IN_CONTEXT':
                    if target_subcontext and line.strip() == target_subcontext:
                        state = 'IN_SUBCONTEXT'

                    elif line.strip() == 'end':
                        # We just finished writing all of the lines for our context.  If
                        # add_line wasn't already in there, write it too.
                        if add_line:
                            if add_line != 'CONTEXT':
                                if add_line in subcontext_lines:
                                    if give_feedback:
                                        print "'%s' configuration already has '%s'" %\
                                            (target_context, add_line)
                                else:
                                    # Create the subcontext
                                    if target_subcontext:
                                        f.write(' ' + target_subcontext + '\n')
                                        f.write('  ' + add_line + '\n')

                                        if use_exit_address_family:
                                            f.write(' exit-address-family\n')
                                    else:
                                        f.write(' ' + add_line + '\n')
                            # If our add_line is CONTEXT, there isn't anything
                            # to do because the context already exists.
                            add_line = None

                        # If we are deleting the entire context, we must also delete the 'end'.
                        # Note that the 'interface' check here is because in frr we cannot
                        # do a "no" on an entire interface. Tracking this via CM-11563.
                        elif del_line == 'CONTEXT' and not target_context.startswith('interface '):
                            state = 'LIMBO'
                            continue

                        state = 'LIMBO'
                # End of case IN_CONTEXT

                elif state == 'IN_SUBCONTEXT':
                    if line.strip().lower() in ['exit-address-family', 'end']:
                        # We just finished writing all of the lines for our subcontext.  If
                        # add_line wasn't already in there, write it too.
                        if add_line and add_line != 'CONTEXT':
                            if add_line in subcontext_lines:
                                if give_feedback:
                                    print "'%s %s' configuration already has '%s'" %\
                                        (target_context, target_subcontext, add_line)
                            else:
                                f.write('  ' + add_line + '\n')
                        add_line = None
                        state = 'IN_CONTEXT'

                    elif line.strip() == 'end':
                        state = 'LIMBO'
                # End of case IN_SUBCONTEXT

                log.debug("{0}: {1}".format(state, line))

                if (not target_context or
                    (not target_subcontext and state == 'IN_CONTEXT') or
                    (target_subcontext and state == 'IN_SUBCONTEXT')):

                    # We are deleting the entire context, so do not write any of the lines.
                    # Note that the 'interface' check here is because in frr we cannot
                    # do a "no" on an entire interface. Tracking this via CM-11563.
                    if del_line == 'CONTEXT' and not line.startswith('interface '):
                        pass

                    elif delete_this_line(line, del_line, del_line_vni, vni, exact_match):
                        found_del_line = True

                    else:
                        # If the user is adding a line to the global context,
                        # do not add the same line multiple times.
                        if not target_context and line == add_line:
                            if give_feedback:
                                print "configuration already has '%s'" % add_line
                            add_line = None
                        f.write(line + '\n')

                    subcontext_lines.append(line.strip())
                else:
                    # If we are deleting the entire ospf context then we also
                    # need to delete the ospf lines under the interfaces
                    if del_line == 'CONTEXT' and target_context == 'router ospf' and line.lstrip().startswith('ip ospf '):
                        pass
                    elif del_line == 'CONTEXT' and target_context == 'router ospf6' and line.lstrip().startswith('ipv6 ospf6 '):
                        pass
                    else:
                        f.write(line + '\n')

                if line.lstrip() == 'exit-vni':
                    vni = None
            # End of "for" loop

            if add_line:
                log.debug("STATE %s, ADD_LINE: %s" % (state, add_line))

                # Note that CONTEXT is a special case.  We hit this if
                # the user is trying to add a context, such as "router bgp 100".
                if add_line == 'CONTEXT':
                    if state not in ('IN_CONTEXT', 'IN_SUBCONTEXT'):
                        f.write('%s\n' % target_context)

                        if target_subcontext:
                            fh.write(' %s\n' % target_subcontext)

                            if use_exit_address_family:
                                fh.write(' exit-address-family\n')
                        f.write('end\n')

                # The context (or sub-context) for which we were looking wasn't in the
                # configuration file.
                else:
                    # Enter the context.  If it is None, use the global context.
                    if target_context:
                        f.write('%s\n' % target_context)

                        # We are in the context.  Enter the sub-context if there is
                        # one. Indent accordingly.
                        if target_subcontext:
                            f.write(' %s\n' % target_subcontext)
                            f.write('  %s\n' % add_line)

                            if use_exit_address_family:
                                f.write(' exit-address-family\n')
                        else:
                            f.write(' %s\n' % add_line)
                        f.write('end\n')
                    else:
                        f.write('%s\n' % add_line)
                        f.write('end\n')

            if vni and line.lstrip() == 'exit-vni':
                vni = None

            # If we deleted lines, we need to truncate the rest of the file so that we
            # do not end up with the lines at the end printed twice.
            f.truncate()

        if give_feedback:
            if not found_del_line and del_line != 'CONTEXT':
                if del_line_vni:
                    print "'%s %s' configuration does not have 'vni %s'" % (target_context, target_subcontext, del_line_vni)
                elif target_subcontext:
                    print "'%s %s' configuration does not have '%s'" % (target_context, target_subcontext, del_line)
                elif target_context:
                    print "'%s' configuration does not have '%s'" % (target_context, del_line)
                else:
                    print "configuration does not have '%s'" % del_line

                if del_line in ('no ptm-enable', 'no ipv6 nd suppress-ra', 'no bgp default ipv4-unicast'):
                    del_line_without_no = del_line.replace('no ', '')
                    print '\nNote that "%s" is on by default so when you do a "net add" NCLU\n' % del_line_without_no +\
                          'will translate that to a "net del" of the "%s" lines in %s' % (del_line, service_name)

        return True

    def bgp_neighbor_sanity(self, context, subcontext, cmd):
        # If we are adding a keyword for a bgp neighbor there are a few sanity
        # checks that we must make
        # - Is the peer defined?
        # - If the peer's name matches an interface is the peer defined as an "interface" peer?
        cmd = cmd.split()

        if context and context.startswith('router bgp') and cmd[0].startswith('neighbor'):
            neighbor = cmd[1]
            keyword = cmd[2]
            interfaces = [x for x in listdir('/sys/class/net/') if x not in ('bonding_masters', 'swid0_eth')]

            # interface peer
            if neighbor in interfaces:
                if keyword != 'interface' and not self.neighbor_interface(neighbor):
                    self.edit_pending_file(context, None,
                                           'neighbor %s interface' % neighbor,
                                           None, True)

            # ipv4/ipv6 peer
            else:
                if keyword not in ('remote-as', 'peer-group') and not self.neighbor_remote_as(neighbor):
                    self.edit_pending_file(context, None,
                                           'neighbor %s remote-as external' % neighbor,
                                           None, True)

    def add(self, context, subcontext, cmd):
        self.bgp_neighbor_sanity(context, subcontext, cmd)
        self.edit_pending_file(context, subcontext, cmd, None, True)

    def delete(self, context, subcontext, cmd, exact_match):

        # 'route-target both' in frr is a macro for configuring 'route-target import'
        # and 'route-target export' so intercept this command and do dels for import
        # and export.
        if 'route-target both' in cmd:
            rt_import = cmd.replace('route-target both', 'route-target import')
            rt_export = cmd.replace('route-target both', 'route-target export')
            self.edit_pending_file(context, subcontext, None, rt_import, exact_match)
            self.edit_pending_file(context, subcontext, None, rt_export, exact_match)
        else:
            self.edit_pending_file(context, subcontext, None, cmd, exact_match)

        # If a deletion occured that erased the last "router bgp ..." configuration info,
        # remove from the pending frr.conf any lines about the route map delay timer.

        # Buffer the pending configuration file.
        if os.path.exists(self.pending_conf):
            with open(self.pending_conf, 'r') as f:
                config_text = f.read()

            # Find any remaining "router bgp ..." information.
            router_info = re.search(r"""^\s*router\s+bgp\s+\d+\s*$""", config_text, re.I | re.M)

            # Delete delay timer related lines, if appropriate.
            if router_info is None:
                config_text = re.sub(r"""^\s*bgp\s+route-map\s+delay-timer.*$""", '', config_text, flags=re.I | re.M)

                with open(self.pending_conf, 'w') as f:
                    f.write(config_text)

    def abort_pending(self):
        for filename in (self.baseline_conf, self.pending_conf, self.pending_baseline_conf,
                         self.baseline_daemons, self.pending_daemons):
            if os.path.isfile(filename):
                os.unlink(filename)

    def commit_pending(self, verbose):
        """
        Make changes persistent by running frr-reload.py.
        """
        files_modified = []
        restart_frr = False
        commit_ok = True

        if (not os.path.isfile(self.pending_baseline_conf) or
            not os.path.isfile(self.pending_conf) or
            not self.init_guts()):
            return (commit_ok, files_modified)

        # If there is a delta, /usr/bin/diff will raise CalledProcessError.
        try:
            check_output(['/usr/bin/diff', self.baseline_daemons, self.pending_daemons])

            # If we are here, there wasn't a change to /etc/frr/daemons.
            try:
                # If frr is enabled but not active, then we need to start frr.
                check_output(['/bin/systemctl', 'is-enabled', service_name])

                try:
                    check_output(['/bin/systemctl', 'is-active', service_name])

                # is-active Exception
                except CalledProcessError as e:
                    restart_frr = True

            # is-enabled Exception
            except CalledProcessError as e:
                pass

        # /usr/bin/diff Exception
        except CalledProcessError as e:
            copyfile(self.pending_daemons, etc_frr_daemons)
            chown_chgroup(etc_frr_daemons, service_name, service_name)

            # A new daemon was enabled in /etc/frr/daemons so we must
            # restart frr so that the new daemon is started. Post CM-11625
            # we can do a start here which will be much less disruptive.
            restart_frr = True

        if restart_frr:

            try:
                output = check_output(['/bin/systemctl', 'enable', service_name])
                if verbose:
                    print output

                output = check_output(['/bin/systemctl', 'reset-failed', service_name])
                if verbose:
                    print output

                output = check_output(['/bin/systemctl', 'restart', service_name])
                if verbose:
                    print output

            except Exception:
                print "ERROR: Failed to restart %s" % service_name

            files_modified.append(etc_frr_daemons)

        os.unlink(self.baseline_daemons)
        os.unlink(self.pending_daemons)

        # If there is a delta, /usr/bin/diff will raise CalledProcessError
        # If there is not a delta to apply do not bother calling frr-reload.py
        try:
            check_output(['/usr/bin/diff', self.pending_baseline_conf, self.pending_conf])
            os.unlink(self.baseline_conf)
            os.unlink(self.pending_baseline_conf)
            os.unlink(self.pending_conf)
            return (commit_ok, files_modified)
        except CalledProcessError as e:
            pass

        try:
            # The /var/run/nclu/frr/ directly is locked down (we want to keep it this way)
            # but /var/run/nclu/ is open for read so cp pending_conf to /var/run/nclu/ so it
            # is in a place where frr-reload.py can read it
            copyfile(self.pending_conf, public_pending_conf)
            copyfile(etc_frr_frr_conf, etc_frr_frr_conf_backup)

            if self.show_linux_command:
                print "Updating %s and applying changes via %s" % (etc_frr_frr_conf, frr_reload_py)

            output = check_output([frr_reload_py,
                                   '--stdout',
                                   '--reload',
                                   public_pending_conf], stderr=STDOUT)
            os.unlink(public_pending_conf)

            if verbose:
                print output

            os.unlink(self.baseline_conf)
            os.unlink(self.pending_baseline_conf)
            os.unlink(self.pending_conf)
            os.unlink(etc_frr_frr_conf_backup)
            files_modified.append(etc_frr_frr_conf)
        except CalledProcessError as e:
            log.exception(e)
            log.error(e.output)
            commit_ok = False
            print "\nERROR: %s failed" % frr_reload_py
            print "\n%s\n%s\n%s\n" % ('=' * 80, e.output, '=' * 80)

            # frr-reload failed to apply all staged changes so we must use
            # frr-reload to undo any changes that did get applied. We do this
            # by reloading the existing /etc/frr/frr.conf
            #
            # It is possible for frr-reload to fail but still partially update
            # frr.conf so replace the current copy with the backup we made
            # prior to calling frr-reload
            copyfile(etc_frr_frr_conf_backup, etc_frr_frr_conf)
            os.unlink(etc_frr_frr_conf_backup)

            try:
                check_output([frr_reload_py,
                              '--stdout',
                              '--reload',
                              etc_frr_frr_conf], stderr=STDOUT)
            except CalledProcessError as e:
                log.exception(e)
                log.error(e.output)
                print "\nERROR: %s failed to restore %s" % (frr_reload_py, etc_frr_frr_conf)
                print "\n%s\n%s\n%s\n" % ('=' * 80, e.output, '=' * 80)

        return (commit_ok, files_modified)

    def show_pending(self, print_errors=True):
        diff_pending = PENDING_COMMIT_NOCHANGE

        if not os.path.isfile(self.pending_conf) and not os.path.isfile(self.pending_daemons):
            return diff_pending

        if not self.init_guts():
            return diff_pending

        if print_diff(self.baseline_daemons, self.pending_daemons, self.color_diffs):
            diff_pending = PENDING_COMMIT_CHANGE
            print """\033[91m\n
********************************************************************
********************************************************************
NOTE: 'net commit' will run 'systemctl restart %s' to apply the
change to %s
********************************************************************
********************************************************************
\033[0m
""" % (service_name, etc_frr_daemons)


        if print_diff(self.pending_baseline_conf, self.pending_conf, self.color_diffs):
            diff_pending = PENDING_COMMIT_CHANGE

        return diff_pending

    def show_config_files(self, user_may_edit):

        if os.path.exists(etc_frr_daemons):
            print "%s" % etc_frr_daemons
            print "==================="

            with open(etc_frr_daemons, 'r') as fh:
                print ''.join(fh.readlines())
            print ''

        if os.path.exists(etc_frr_frr_conf):
            print "%s" % etc_frr_frr_conf
            print "======================="

            # If the user has edit privs then do not bother hiding passwords
            if user_may_edit:
                with open(etc_frr_frr_conf, 'r') as fh:
                    print ''.join(fh.readlines())

            # If they can only do "show" commands then replace any passwords with PASSWORD_REMOVED
            else:
                output = []
                with open(etc_frr_frr_conf, 'r') as fh:
                    for line in fh.readlines():

                        if 'authentication-key' in line:
                            re_ospf_auth_key = re.search('^(.* ospf authentication-key\s+)\S+(.*)$', line)
                            if re_ospf_auth_key:
                                line = re_ospf_auth_key.group(1) + 'PASSWORD_REMOVED' + re_ospf_auth_key.group(2) + '\n'
                                output.append(line)
                            else:
                                output.append(line)

                        elif 'password' in line and 'password 8 ' not in line:
                            re_password = re.search('^(.*password\s+)\S+(.*)$', line)

                            if re_password:
                                line = re_password.group(1) + 'PASSWORD_REMOVED' + re_password.group(2) + '\n'
                                output.append(line)
                            else:
                                output.append(line)
                        else:
                            output.append(line)
                print ''.join(output)
            print ''

    def show_config_commands(self, user_may_edit):
        interface_commands = []
        commands = []
        supported = []
        not_supported = []

        if os.path.exists(etc_frr_frr_conf):
            try:
                frr_conf = frr_reload.Config()
                frr_conf.load_from_file(etc_frr_frr_conf)
            except Exception as e:
                msg = frr_reload_exception_to_string(etc_frr_frr_conf, e)
                log.error(msg)
                print msg
                return (supported, not_supported)

            for context in frr_conf.contexts.itervalues():

                if len(context.keys) == 1:
                    key = context.keys[0]

                    if key.startswith('router bgp'):
                        printf_alt = '%s\\n' % key
                        asn = key.split()[2]

                        if 'vrf' in key:
                            # router bgp 100 vrf RED
                            vrf = key.split()[4]
                        else:
                            vrf = None

                        if vrf:
                            commands.append((printf_alt, 'net add bgp vrf %s autonomous-system %s' % (vrf, asn)))
                        else:
                            commands.append((printf_alt, 'net add bgp autonomous-system %s' % asn))

                        for line in context.lines:
                            printf_alt = '%s\\n  %s\\n' % (key, line)
                            cmd = 'net add bgp %s' % line
                            cmd = cmd.replace(' bgp bgp ', ' bgp ')

                            if vrf:
                                cmd = cmd.replace("net add bgp", "net add bgp vrf %s" % vrf)

                            if 'no bgp default ipv4-unicast' in cmd:
                                cmd = cmd.replace('no bgp default ipv4-unicast', 'default ipv4-unicast')
                                cmd = cmd.replace('net add', 'net del')
                            elif 'timers bgp' in cmd:
                                cmd = cmd.replace('timers bgp', 'timers')

                            commands.append((printf_alt, cmd))

                    elif key.startswith('router ospf6'):
                        for line in context.lines:
                            printf_alt = '%s\\n  %s\\n' % (key, line)
                            commands.append((printf_alt, 'net add ospf6 %s' % line))

                    elif key.startswith('router ospf'):
                        for line in context.lines:
                            printf_alt = '%s\\n  %s\\n' % (key, line)
                            cmd = 'net add ospf %s' % line
                            cmd = cmd.replace(' ospf ospf ', ' ospf ')
                            commands.append((printf_alt, cmd))

                    elif key.startswith('vrf'):
                        for line in context.lines:
                            printf_alt = '%s\\n  %s\\n' % (key, line)

                            # vrf BLUE
                            #  ip pim rp 1.1.1.1 224.0.0.0/4
                            #
                            # translates to "net add vrf BLUE ip pim rp 1.1.1.1 224.0.0.0/4"
                            if line.startswith('ip pim'):
                                pim_options = ' '.join(line.split()[2:])
                                cmd = 'net add pim %s %s' % (key, pim_options)
                            else:
                                cmd = 'net add %s %s' % (key, line)

                            commands.append((printf_alt, cmd))

                    elif context.lines:
                        if key.startswith('interface'):
                            for line in context.lines:
                                if line == 'no ipv6 nd suppress-ra':
                                    interface_commands.append('net del %s ipv6 nd suppress-ra' % key)

                                elif line == 'no ptm-enable':
                                    interface_commands.append('net del %s ptm-enable' % key)

                                elif line.startswith('ip ospf'):
                                    line = line.replace('ip ospf', 'ospf')
                                    interface_commands.append('net add %s %s' % (key, line))

                                elif line.startswith('ipv6 ospf6'):
                                    line = line.replace('ipv6 ospf6', 'ospf6')
                                    interface_commands.append('net add %s %s' % (key, line))

                                elif line.startswith('ip igmp'):
                                    line = line.replace('ip igmp', 'igmp')
                                    interface_commands.append('net add %s %s' % (key, line))

                                elif line.startswith('ip pim'):
                                    line = line.replace('ip pim', 'pim')
                                    interface_commands.append('net add %s %s' % (key, line))

                                elif line.startswith('ip msdp'):
                                    line = line.replace('ip msdp', 'msdp')
                                    interface_commands.append('net add %s %s' % (key, line))

                                else:
                                    interface_commands.append('net add %s %s' % (key, line))
                        else:
                            if key.startswith('route-map'):
                                key = key.replace('route-map', 'routing route-map')
                            elif key.startswith('line vty'):
                                key = key.replace('line vty', 'routing line vty')

                            for line in context.lines:
                                printf_alt = '%s\\n  %s\\n' % (key, line)
                                commands.append((printf_alt, 'net add %s %s' % (key, line)))
                    else:
                        printf_alt = '%s\\n' % key

                        # babelfish
                        if key.startswith('log ') or key.startswith('ptm-enable'):
                            key = 'routing ' + key

                        elif key.startswith('ip community-list'):
                            key = key.replace('ip community-list', 'routing community-list')

                        elif key.startswith('ip extcommunity-list'):
                            key = key.replace('ip extcommunity-list', 'routing extcommunity-list')

                        elif key.startswith('ip as-path'):
                            key = key.replace('ip as-path', 'routing as-path')

                        elif key.startswith('ip multicast-routing'):
                            key = key.replace('ip multicast-routing', 'routing multicast-routing')

                        elif key.startswith('ip pim '):
                            key = key.replace('ip pim', 'pim')

                        elif key.startswith('ip msdp '):
                            key = key.replace('ip msdp', 'msdp')

                        elif key.startswith('ip prefix-list'):
                            key = key.replace('ip prefix-list', 'routing prefix-list ipv4')

                        elif key.startswith('ipv6 prefix-list'):
                            key = key.replace('ipv6 prefix-list', 'routing prefix-list ipv6')

                        elif key.startswith('route-map'):
                            key = key.replace('route-map', 'routing route-map')

                        elif key.startswith('ip protocol'):
                            key = key.replace('ip protocol', 'routing protocol')

                        elif key.startswith('ip import-table'):
                            key = key.replace('ip import-table', 'routing import-table')

                        elif key.startswith('ip route'):
                            key = key.replace('ip route', 'routing route')

                        elif key.startswith('ip mroute'):
                            key = key.replace('ip mroute', 'routing mroute')

                        elif key.startswith('ipv6 route'):
                            key = key.replace('ipv6 route', 'routing route')

                        elif key.startswith('service'):
                            key = key.replace('service', 'routing service')

                        elif key.startswith('enable password'):
                            key = key.replace('enable password', 'routing enable password')

                        elif key.startswith('agentx'):
                            key = key.replace('agentx', 'routing agentx')

                        elif key.startswith('password'):
                            key = key.replace('password', 'routing password')

                        elif key.startswith('debug bgp'):
                            key = key.replace('debug bgp', 'bgp debug')

                        elif key.startswith('debug msdp'):
                            key = key.replace('debug msdp', 'msdp debug')

                        elif key.startswith('debug ospf6'):
                            key = key.replace('debug ospf6', 'ospf6 debug')

                        elif key.startswith('debug ospf'):
                            key = key.replace('debug ospf', 'ospf debug')

                        elif key.startswith('debug pim'):
                            key = key.replace('debug pim', 'pim debug')

                        elif key.startswith('debug igmp'):
                            key = key.replace('debug igmp', 'igmp debug')

                        elif key.startswith('debug zebra'):
                            key = key.replace('debug zebra', 'routing zebra debug')

                        elif key.startswith('debug mroute'):
                            key = key.replace('debug mroute', 'routing mroute debug')

                        cmd = 'net add %s' % key
                        commands.append((printf_alt, 'net add %s' % key))

                elif len(context.keys) >= 2:
                    key = context.keys[0]
                    subkey = context.keys[1]

                    if key.startswith('router bgp'):

                        if 'vrf' in key:
                            # router bgp 100 vrf RED
                            vrf = key.split()[4]
                        else:
                            vrf = None

                        if subkey.startswith('address-family'):
                            (_, afi, safi) = subkey.strip().split()

                            if subkey == 'address-family l2vpn evpn':
                                if len(context.keys) == 3:
                                    vni = 'evpn %s ' % context.keys[2]
                                else:
                                    vni = 'evpn '
                            else:
                                vni = ''

                            for line in context.lines:
                                printf_alt = '%s\\n  %s\\n  %s\\n' % (key, subkey, line)
                                add_del = 'add'

                                if re.search('no neighbor \S+ activate', line):
                                    line = line.replace('no neighbor', 'neighbor')
                                    add_del = 'del'

                                if vrf:
                                    commands.append((printf_alt, 'net %s bgp vrf %s %s %s %s%s' % (add_del, vrf, afi, safi, vni, line)))
                                else:
                                    commands.append((printf_alt, 'net %s bgp %s %s %s%s' % (add_del, afi, safi, vni, line)))

                        else:
                            log.info("%s show_config_commands() unsupported BGP subkey %s" % (service_name, subkey))
                    else:
                        log.info("%s show_config_commands() unsupported key %s" % (service_name, pformat(context.keys)))
                else:
                    log.info("%s show_config_commands() unsupported key %s" % (service_name, pformat(context.keys)))


        # Create a temporary NetworkDocopt object so we can see which 'commands' are supported.
        # We need to replace the 'frr' in our __doc__ with 'net'.
        cli = NetworkDocopt(__doc__.replace('    frr ', '    net '))

        interface_commands = globify_interface_commands(interface_commands)
        for cmd in interface_commands:

            # We do not want the user to see the "Command not found"
            # error if this is an unsupported command so use Capturing
            # but do nothing with the output.
            with Capturing() as _:
                cli.run(cmd.split())

            if cli.match:
                supported.append(cmd)
            else:
                cmd = cmd.split()
                ifname_glob = cmd[3]
                ifnames = []
                cmd = ' '.join(cmd[4:])

                if ifname_is_glob(ifname_glob):
                    ifnames = ifname_expand_glob(ifname_glob)
                else:
                    ifnames.append(ifname_glob)

                for ifname in ifnames:
                    log.info("%s interface command '%s' has not been wrappered" % (service_name, cmd))
                    echo_cmd = "printf 'interface %s\\n%s\\n' >> %s" % (ifname, cmd, etc_frr_frr_conf)
                    not_supported.append(echo_cmd)
            cli.re_init()

        for protocol_commands in (False, True):

            for (printf_alt, cmd) in commands:

                if protocol_commands:
                    if cmd.startswith('net add routing'):
                        continue
                else:
                    if not cmd.startswith('net add routing'):
                        continue

                if cmd == '#':
                    supported.append(cmd)
                    continue

                # We do not want the user to see the "Command not found"
                # error if this is an unsupported command so use Capturing
                # but do nothing with the output.
                with Capturing() as _:
                    cli.run(cmd.split())

                if cli.match:
                    supported.append(cmd)
                else:
                    log.info("%s protocol command '%s' has not been wrappered" % (service_name, cmd))
                    echo_cmd = "printf '%s' >> %s" % (printf_alt, etc_frr_frr_conf)
                    not_supported.append(echo_cmd)
                cli.re_init()

        return (supported, not_supported)

    def show_config_summary(self, user_may_edit, summary):

        # Nothing to do so return
        if not os.path.exists(etc_frr_frr_conf):
            return

        try:
            frr_conf = frr_reload.Config()
            frr_conf.load_from_file(etc_frr_frr_conf)

        except Exception as e:
            msg = frr_reload_exception_to_string(etc_frr_frr_conf, e)
            log.error(msg)
            print msg
            return

        for context in frr_conf.contexts.itervalues():

            if len(context.keys) == 1:
                key = context.keys[0]

                if key.startswith('interface '):
                    key = key.replace('interface ', 'INTERFACE:')

                if key not in summary:
                    summary[key] = OrderedDict()

                if context.lines:
                    if 'COMMAND_LIST' not in summary[key]:
                        summary[key]['COMMAND_LIST'] = []
                    summary[key]['COMMAND_LIST'].extend(context.lines)

                    if key.startswith('INTERFACE:'):
                        summary[key]['COMMAND_LIST'] = sort_for_humans(summary[key]['COMMAND_LIST'])

            elif len(context.keys) >= 2:
                key = context.keys[0]
                subkey = context.keys[1]

                if key not in summary:
                    summary[key] = OrderedDict()

                if len(context.keys) == 2:
                    if subkey not in summary[key]:
                        summary[key][subkey] = OrderedDict()
                    summary[key][subkey]['COMMAND_LIST'] = context.lines

                # "address-family l2vpn evpn" "vni 10" is the only place we have this many keys
                elif len(context.keys) == 3:
                    subsubkey = context.keys[2]
                    if subkey not in summary[key]:
                        summary[key][subkey] = OrderedDict()

                    if subsubkey not in summary[key][subkey]:
                        summary[key][subkey][subsubkey] = []
                    summary[key][subkey][subsubkey] = context.lines

                else:
                    log.info("%s show_config_commands() unsupported key %s" % (service_name, pformat(context.keys)))

            else:
                log.info("%s show_config_commands() unsupported key %s" % (service_name, pformat(context.keys)))


    def ip_route_show(self, vtysh_args, zebra_output):

        def get_routing_entry_for():
            if zebra_output:
                for line in zebra_output.splitlines():
                    if line.startswith('Routing entry for'):
                        return line.strip().split()[3]
            return None

        # show ip route x.x.x.x/24
        # ip route show x.x.x.x/24
        #
        # show ip route x.x.x.x
        # ip route show
        if vtysh_args.startswith('show ip route'):
            re_prefix_prefixlen = re_search('^show ip route.* (\d+\.\d+\.\d+\.\d+\/\d+)$', vtysh_args)
            re_prefix = re_search('^show ip route.* (\d+\.\d+\.\d+\.\d+)$', vtysh_args)
            use_ipv6 = False

        elif vtysh_args.startswith('show ipv6 route'):
            re_prefix_prefixlen = re_search('^show ipv6 route.* ([0-9a-fA-F]\S+\/\d+)$', vtysh_args)
            re_prefix = re_search('^show ipv6 route.* ([0-9a-fA-F]\S+:\S+)$', vtysh_args)
            use_ipv6 = True

        else:
            re_prefix_prefixlen = None
            re_prefix = None
            use_ipv6 = False

        if re_prefix_prefixlen:
            ip = re_prefix_prefixlen.group(1)

        elif re_prefix:
            ip = re_prefix.group(1)
            ip = get_routing_entry_for()
        else:
            ip = None

        if ip:
            re_vrf = re.search('vrf (\S+)', vtysh_args)
            vrf = re_vrf.group(1) if re_vrf else None

            header = "FIB entry for %s" % vtysh_args.split()[-1]
            if vrf:
                header += " in vrf %s" % vrf

            print "%s\n%s" % (header, '=' * len(header))

            cmd = ['/sbin/ip',]

            if use_ipv6:
                cmd.append('-6')

            cmd.append('route')
            cmd.append('show')

            if vrf:
                cmd.append('vrf')
                cmd.append(vrf)

            cmd.append(ip)

            if self.show_linux_command:
                print "exec: ''" % ' '.join(cmd)

            try:
                print check_output(cmd)
            except Exception:
                pass

    def del_interfaces(self, ifname):

        # Nothing to do so return
        if not os.path.exists(etc_frr_frr_conf):
            return

        # Create a list of interface names.
        if isinstance(ifname, list):
            ifnames = map(str, ifname)  # The elements of the list are probably already strings.
        elif isinstance(ifname, str):
            if ifname_is_glob(ifname):
                ifnames = ifname_expand_glob(ifname)
            elif ifname_is_valid(ifname) and not ifname_is_globish(ifname):
                ifnames = [ifname]
        else:
            ifnames = []

        try:
            frr_conf = frr_reload.Config()
            frr_conf.load_from_file(etc_frr_frr_conf)

        except Exception as e:
            msg = frr_reload_exception_to_string(etc_frr_frr_conf, e)
            log.error(msg)
            print msg
            return

        for ifname in ifnames:
            context_to_del = 'interface %s' % ifname
            subcontext_to_del = None
            cmd = 'CONTEXT'
            self.delete(context_to_del, subcontext_to_del, cmd, True)

            # Delete any sub-interfaces for this interface
            sub_interface_base = context_to_del + '.'

            for context in frr_conf.contexts.keys():
                if context[0].startswith(sub_interface_base):
                    self.delete(context[0], None, cmd, True)

    def exec_vtysh_show_or_clear(self, args, vtysh_args):
        # In frr, bgp allows the user to do "show ip bgp" or "show bgp"
        # but ospf only allows "show ip ospf".  This is a bit awkward for
        # the user if they are working with both so we made "show ospf" but
        # we need to change that to "show ip ospf"
        vtysh_args = vtysh_args.replace('show ospf6', 'show ipv6 ospf6')
        vtysh_args = vtysh_args.replace('show ospf', 'show ip ospf')
        vtysh_args = vtysh_args.replace('show igmp', 'show ip igmp')
        vtysh_args = vtysh_args.replace('show mroute', 'show ip mroute')
        vtysh_args = vtysh_args.replace('show pim', 'show ip pim')
        vtysh_args = vtysh_args.replace('show msdp', 'show ip msdp')

        vtysh_args = vtysh_args.replace('clear ospf', 'clear ip ospf')
        vtysh_args = vtysh_args.replace('clear bgp', 'clear ip bgp')
        vtysh_args_list = []
        bgp_summary_combined = False
        bgp_vrf_summary_combined = None

        if vtysh_args.startswith('show bgp'):
            re_show_bgp_vrf_summary = re.search('show bgp vrf (\S+) summary', vtysh_args)
            re_show_bgp_vrf_neighbor = re.search('show bgp vrf (\S+) neighbor', vtysh_args)

            # Have this dump both IPv4 and IPv6 tables
            if vtysh_args == 'show bgp':
                vtysh_args_list.append('show bgp ipv4 unicast')
                vtysh_args_list.append('show bgp ipv6 unicast')

            # Have this dump both IPv4 and IPv6 summary
            elif vtysh_args == 'show bgp summary':
                vtysh_args_list.append('show bgp ipv4 unicast summary')
                vtysh_args_list.append('show bgp ipv6 unicast summary')
                vtysh_args_list.append('show bgp l2vpn evpn summary')

            elif vtysh_args == 'show bgp summary json':
                bgp_summary_combined = True

            elif vtysh_args == 'show bgp update-groups':
                vtysh_args_list.append('show bgp ipv4 unicast update-groups')
                vtysh_args_list.append('show bgp ipv6 unicast update-groups')

            elif ((vtysh_args.startswith('show bgp neighbor') or re_show_bgp_vrf_neighbor) and
                  ('advertised-routes' in vtysh_args or 'received-routes' in vtysh_args or 'routes' in vtysh_args)):
                vtysh_args = vtysh_args.replace('neighbor', 'ipv4 unicast neighbor')
                vtysh_args_list.append(vtysh_args)

            # Nothing special for these, just run as they are
            elif (vtysh_args.startswith('show bgp ipv4 unicast') or
                  vtysh_args.startswith('show bgp ipv4 labeled-unicast') or
                  vtysh_args.startswith('show bgp ipv6 unicast') or
                  vtysh_args.startswith('show bgp neighbor')):
                vtysh_args_list.append(vtysh_args)

            # Nothing special for show bgp evpn commands, just run as they are
            elif (vtysh_args.startswith('show bgp l2vpn evpn')):
                vtysh_args_list.append(vtysh_args)

            elif re_show_bgp_vrf_summary:
                vrf = re_show_bgp_vrf_summary.group(1)

                if 'json' in vtysh_args:
                    bgp_vrf_summary_combined = vrf
                else:
                    vtysh_args_list.append('show bgp vrf %s ipv4 unicast summary' % vrf)
                    vtysh_args_list.append('show bgp vrf %s ipv6 unicast summary' % vrf)

            # If we are here then we are running this command:
            # frr show bgp (<ipv4>|<ipv4/prefixlen>|<ipv6>|<ipv6/prefixlen>) [bestpath|multipath] [json]
            #
            # So we must look to see if we are dealing with an IPv4 prefix or an IPv6 prefix
            # and add 'ipv4 unicast' or 'ipv6 unicast' to the command
            else:

                # For now vrf is ipv4 only but you can only do
                if args.get('vrf'):
                    if args.get('<ipv4>') or args.get('<ipv4/prefixlen>'):
                        vtysh_args = vtysh_args.replace('show bgp', 'show ip bgp')

                elif args.get('<ipv4>') or args.get('<ipv4/prefixlen>'):
                    vtysh_args = vtysh_args.replace('show bgp', 'show bgp ipv4 unicast')

                elif args.get('<ipv6>') or args.get('<ipv6/prefixlen>'):
                    vtysh_args = vtysh_args.replace('show bgp', 'show bgp ipv6 unicast')

                vtysh_args_list.append(vtysh_args)

        # frr show route [vrf <text>] [<ipv4>|<ipv4/prefixlen>|<ipv6>|<ipv6/prefixlen>|bgp|connected|ospf|static|summary]
        # frr show route [vrf <text>] (<ipv4/prefixlen>|<ipv6/prefixlen>) longer-prefixes
        elif vtysh_args.startswith('show route') and not vtysh_args.startswith('show route-map'):
            re_show_route_vrf = re.search('^show route vrf (\S+)(.*)', vtysh_args)
            vrf = re_show_route_vrf.group(1) if re_show_route_vrf else None

            if vtysh_args == 'show route summary':
                vtysh_args_list.append('show ip route summary')
                vtysh_args_list.append('show ipv6 route summary')

            elif vtysh_args == 'show route':
                vtysh_args_list.append('show ip route')
                vtysh_args_list.append('show ipv6 route')

            elif re_show_route_vrf and not (args.get('<ipv4>') or args.get('<ipv4/prefixlen>') or args.get('<ipv6>') or args.get('<ipv6/prefixlen>')):
                show_options = re_show_route_vrf.group(2).strip()
                vtysh_args_list.append('show ip route vrf %s %s' % (vrf, show_options))
                vtysh_args_list.append('show ipv6 route vrf %s %s' % (vrf, show_options))

            else:
                if args.get('<ipv6>') or args.get('<ipv6/prefixlen>') or args.get('ospf6') or args.get('ripng') or args.get('ipv6'):
                    vtysh_args = vtysh_args.replace('show route', 'show ipv6 route')

                else:
                    vtysh_args = vtysh_args.replace('show route', 'show ip route')

                if 'json' not in vtysh_args:
                    # When ip_route_show() is called we will print the kernel entry for this route
                    # so print a little header here so the user can tell which is which
                    header = "RIB entry for %s" % vtysh_args.split()[-1]
                    if vrf:
                        header += " in vrf %s" % vrf
                    print "%s\n%s" % (header, '=' * len(header))
                vtysh_args_list.append(vtysh_args)

        # Change "clear ip bgp all" to "clear ip bgp *"
        elif 'clear' in vtysh_args and 'bgp' in vtysh_args and 'all' in vtysh_args:
            vtysh_args_list.append(vtysh_args.replace(' all', ' *'))

        elif vtysh_args == 'show debugs':
            vtysh_args_list.append('show debugging bgp')
            vtysh_args_list.append('show debugging mpls ldp')
            vtysh_args_list.append('show debugging ospf')
            vtysh_args_list.append('show debugging pim')
            vtysh_args_list.append('show debugging zebra')

        else:
            vtysh_args_list.append(vtysh_args)

        if bgp_vrf_summary_combined:
            vrf = bgp_vrf_summary_combined

            try:
                ipv4_output = check_output(['/usr/bin/vtysh', '-c', 'show ip bgp vrf %s summary json' % vrf])
                ipv6_output = check_output(['/usr/bin/vtysh', '-c', 'show bgp vrf %s summary json' % vrf])

                if 'No such BGP instance exist' in ipv4_output:
                    ipv4_output = '{}'

                if 'No such BGP instance exist' in ipv6_output:
                    ipv6_output = '{}'

                print """{
"ipv4 unicast" : %s,
"ipv6 unicast" : %s
}
""" % (ipv4_output, ipv6_output)

            except Exception as e:
                log.debug("bgp_summary_combined returned with error:\n%s\n" % e)
                print "{}"

        elif bgp_summary_combined:
            try:
                ipv4_output = check_output(['/usr/bin/vtysh', '-c', 'show bgp ipv4 unicast summary json'])
                ipv6_output = check_output(['/usr/bin/vtysh', '-c', 'show bgp ipv6 unicast summary json'])
                evpn_output = check_output(['/usr/bin/vtysh', '-c', 'show bgp l2vpn evpn summary json'])
                print """{
"ipv4 unicast" : %s,
"ipv6 unicast" : %s,
"l2vpn evpn" : %s
}
""" % (ipv4_output, ipv6_output, evpn_output)
            except Exception as e:
                log.info("bgp_summary_combined returned with error:\n%s\n" % e)
                print "{}"

        else:
            print_header = True if len(vtysh_args_list) > 1 else False

            for vtysh_args in vtysh_args_list:
                cmd = ['/usr/bin/vtysh', '-c', vtysh_args]

                if self.show_linux_command and 'json' not in vtysh_args:
                    print "exec 'vtysh -c \"%s\"'\n" % vtysh_args

                if print_header:
                    print "\n%s\n%s" % (vtysh_args, '=' * len(vtysh_args))

                try:
                    output = check_output(cmd, stderr=STDOUT)
                except Exception as e:
                    if vtysh_args.startswith('show ip igmp'):
                        output = "ERROR: frr must be enabled for 'net show igmp' commands"
                    elif 'show debugging mpls' not in vtysh_args:
                        log.debug("'%s' returned with error:\n%s\n" % (' '.join(cmd), e))
                        output = None
                    else:
                        output = None

                if output:
                    print output

                # For "show route" also show the kernel route
                if vtysh_args.startswith('show ip route') or vtysh_args.startswith('show ipv6 route'):
                    self.ip_route_show(vtysh_args, output)

    def eval_command_line_args(self, cli):
        """
        NetworkDocopt will populate the self.args dictionary based on the command
        line args entered by the user.
        """
        command_ok = True
        args = cli.args

        # If the user does "net show route" when frr isn't running, don't print
        # an error.  Just return blank.
        if not (args.get('show') or args.get('clear')):
            if not self.init_guts():
                return False

        bgp_bgp = ['always-compare-med',
                   'bestpath',
                   'client-to-client reflection',
                   'cluster-id',
                   'confederation peers',
                   'default ipv4-unicast',
                   'default local-preference',
                   'default show-hostname',
                   'default subgroup-pkt-queue-max',
                   'deterministic-med',
                   'disable-ebgp-connected-route-check',
                   'enforce-first-as',
                   'fast-external-failover',
                   'graceful-restart',
                   'listen',
                   'log-neighbor-changes',
                   'max-med',
                   'network import-check',
                   'route-map delay-timer',
                   'route-reflector allow-outbound-policy',
                   'router-id']
        ospf_ospf = ['router-id', ]
        ifnames = []
        bgppeers = []

        # If <interface> is a glob, then populate ifnames with all of the interface
        # names produced by that glob.  Otherwise, populate ifnames with <interface>.
        if args.get('<interface>'):
            ifname = args.get('<interface>')

            if ifname_is_glob(str(ifname)):
                for x in ifname_expand_glob(str(ifname)):
                    ifnames.append(x)

            elif ifname_is_valid(str(ifname)) and not ifname_is_globish(str(ifname)):
                ifnames = [ifname, ]

            else:
                print "ERROR: A valid interface name or glob is needed."
                return False

        elif args.get('vlan') and args.get('<number-range-list>'):
            vlan_numbers = glob_to_numbers(str(args.get('<number-range-list>')))

            for number in vlan_numbers:
                ifnames.append("vlan%d" % number)

        elif args.get('<bgppeer>'):
            bgppeer = str(args.get('<bgppeer>'))

            if ifname_is_glob(bgppeer):
                for x in ifname_expand_glob(bgppeer):
                    bgppeers.append(x)
            else:
                bgppeers.append(bgppeer)

        # Sanity check that the log file specified is in a place where
        # frr has write permissions
        if args.get('add') and args.get('log') and args.get('file'):
            filename = args.get('<text>')

            if not filename.startswith('/var/log/%s/' % service_name):
                print "ERROR: frr log files must be in /var/log/%s/" % service_name
                return False

        if args.get('show') or args.get('clear'):
            cmd = []

            for node in cli.match[1:]:
                if node.value:
                    if node.value.text == '<interface>':
                        cmd.append(node.value.text)
                    elif node.value.text.startswith('<'):
                        cmd.append(node.value.value)
                    else:
                        cmd.append(node.value.text)

                # Default afi to 'ipv4' and safi to 'unicast'
                else:
                    if str(node) == '[ipv4|ipv6]':
                        cmd.append('ipv4')
                    elif str(node) == '[unicast|multicast]':
                        cmd.append('unicast')
                    elif str(node) == '[ipv4 unicast|ipv6 unicast]':
                        cmd.append('ipv4')
                        cmd.append('unicast')

            vtysh_args = ' '.join(map(str, cmd))

            # Support show commands with globs of interfaces
            # net show ospf interface swp3-4
            if ifnames:
                for ifname in ifnames:
                    vtysh_args_for_ifname = vtysh_args.replace('<interface>', ifname)
                    self.exec_vtysh_show_or_clear(args, vtysh_args_for_ifname)
            elif bgppeers:
                for peer in bgppeers:
                    vtysh_args_for_bgppeer = vtysh_args.replace('<bgppeer>', peer)
                    self.exec_vtysh_show_or_clear(args, vtysh_args_for_bgppeer)
            else:
                self.exec_vtysh_show_or_clear(args, vtysh_args)

        elif args.get('add') or args.get('del'):
            contexts = []
            subcontext = None

            # (add|del) will be the node at #1
            # entry #2 will be the bgp, ospf, interface, etc node
            add_del_node = cli.match[1]
            context_node = cli.match[2]

            # Set node_cmd_start to 3 so we don't look at "(add|del) bgp" when
            # building the actual frr cmd string.
            node_cmd_start = 3

            cmd_is_CONTEXT = False

            # If the user entered "bgp autonomous-system 100",
            # we must translate that to "router bgp 100".
            if context_node.value.text == 'bgp' and cli.match[3].value:
                keyword = cli.match[3].value.text

                if keyword == 'autonomous-system':
                    cli.match[2].value.value = 'router'
                    cli.match[2].value.value_raw = 'router'
                    cli.match[3].value.value = 'bgp'
                    cli.match[3].value.value_raw = 'bgp'
                    context_node = cli.match[3]
                    cmd_is_CONTEXT = True

                # net add bgp    vrf BLUE autonomous-system 77
                # net add router bgp 77   vrf               BLUE
                # 0   1   2      3   4    5                 6
                elif keyword == 'vrf':
                    if len(cli.match) >= 7:
                        if cli.match[5].value.text == 'autonomous-system':
                            vrf = cli.match[4].value.value
                            asn = cli.match[6].value.value
                            cli.match[2].value.value = 'router'
                            cli.match[2].value.value_raw = 'router'
                            cli.match[3].value.value = 'bgp'
                            cli.match[3].value.value_raw = 'bgp'
                            cli.match[4].value.value = asn
                            cli.match[4].value.value_raw = asn
                            cli.match[5].value.value = 'vrf'
                            cli.match[5].value.value_raw = 'vrf'
                            cli.match[6].value.value = vrf
                            cli.match[6].value.value_raw = vrf
                            context_node = cli.match[3]
                            cmd_is_CONTEXT = True

            # net del ospf
            # net del ospf6
            elif context_node.value.text in ('ospf', 'ospf6') and len(cli.match) == 3:
                cmd_is_CONTEXT = True

            elif context_node.value.text == 'bond' or context_node.value.text == 'loopback':
                cli.match[2].value.value = 'interface'
                cli.match[2].value.value_raw = 'interface'

            # Find the context and subcontext.
            if args.get('debug'):
                # debugs are at the top level in frr but in nclu we put them under
                # the protocol...so 'net add bgp debug bestpath'. We configure this in
                # frr at the top level and as "debug bgp bestpath".

                # convert "routing zebra debug events" to "debug zebra events"
                if args.get('routing') and args.get('zebra'):
                    cli.match[3].value.value = 'debug'
                    cli.match[3].value.value_raw = 'debug'
                    cli.match[4].value.value = 'zebra'
                    cli.match[4].value.value_raw = 'zebra'
                    node_cmd_start = 3
                    contexts.append(None)

                # convert "routing mroute debug detail" to "debug mroute detail"
                elif args.get('routing') and args.get('mroute'):
                    cli.match[3].value.value = 'debug'
                    cli.match[3].value.value_raw = 'debug'
                    cli.match[4].value.value = 'mroute'
                    cli.match[4].value.value_raw = 'mroute'
                    node_cmd_start = 3
                    contexts.append(None)

                else:
                    # flip 'bgp debug' to 'debug bgp'
                    cli.match[3].value.value = cli.match[2].value.value
                    cli.match[3].value.value_raw = cli.match[2].value.value_raw
                    cli.match[2].value.value = 'debug'
                    cli.match[2].value.value_raw = 'debug'
                    node_cmd_start = 2
                    contexts.append(None)

            # convert 'routing protocol' to 'ip protocol', etc
            elif (args.get('routing') and
                  (args.get('protocol') or (args.get('route') and args.get('<ipv4/prefixlen>') or
                   (args.get('as-path') and args.get('access-list')) or
                   args.get('import-table') or
                   args.get('multicast-routing') or
                   args.get('mroute') or
                   args.get('community-list') or
                   args.get('extcommunity-list')))):
                cli.match[2].value.value = 'ip'
                cli.match[2].value.value_raw = 'ip'

                if args.get('route') and args.get('<ipv4/prefixlen>'):
                    sanity_check_ip_host_vs_subnet(args, cli)

                node_cmd_start = 2
                contexts.append(None)

            # convert 'routing route <ipv6/prefixlen>' to 'ipv6 route'
            elif args.get('routing') and args.get('route') and args.get('<ipv6/prefixlen>'):
                cli.match[2].value.value = 'ipv6'
                cli.match[2].value.value_raw = 'ipv6'
                node_cmd_start = 2
                contexts.append(None)
                sanity_check_ip_host_vs_subnet(args, cli)

            # convert 'routing route-map ...' to 'route-map ...'
            elif args.get('routing') and args.get('route-map'):
                cli.match[2].value.value = ''
                cli.match[2].value.value_raw = ''
                context_node = cli.match[3]
                contexts.append(' '.join(map(str, cli.argv_expanded[3:7])))
                node_cmd_start = 7

                # 'net add routing route-map FOO permit 10' is 7 words
                if len(cli.argv_expanded) == 7:
                    cmd_is_CONTEXT = True

            elif args.get('line') and args.get('vty'):
                cli.match[2].value.value = ''
                cli.match[2].value.value_raw = ''
                context_node = cli.match[4]
                node_cmd_start = 5
                contexts.append(' '.join(map(str, cli.argv_expanded[3:5])))

                # 'net add routing line vty' is 5 words
                if len(cli.argv_expanded) == 5:
                    cmd_is_CONTEXT = True

            # convert 'routing log' to just 'log', etc
            elif (args.get('routing') and
                  (args.get('log') or
                   args.get('service') or
                   args.get('agentx') or
                   args.get('password') or
                   args.get('ptm-enable'))):
                cli.match[2].value.value = ''
                cli.match[2].value.value_raw = ''
                node_cmd_start = 2
                contexts.append(None)
                context_node = cli.match[3]

            # convert 'routing prefix-list ipv4' to 'ip prefix-list'
            elif args.get('routing') and args.get('prefix-list') and args.get('ipv4'):
                cli.match[2].value.value = ''
                cli.match[2].value.value_raw = ''
                cli.match[3].value.value = 'ip'
                cli.match[3].value.value_raw = 'ip'
                cli.match[4].value.value = 'prefix-list'
                cli.match[4].value.value_raw = 'prefix-list'
                node_cmd_start = 2
                contexts.append(None)
                sanity_check_ip_host_vs_subnet(args, cli)

            # convert 'routing prefix-list ipv6' to 'ipv6 prefix-list'
            elif args.get('routing') and args.get('prefix-list') and args.get('ipv6'):
                cli.match[2].value.value = ''
                cli.match[2].value.value_raw = ''
                cli.match[3].value.value = 'ipv6'
                cli.match[3].value.value_raw = 'ipv6'
                cli.match[4].value.value = 'prefix-list'
                cli.match[4].value.value_raw = 'prefix-list'
                node_cmd_start = 2
                contexts.append(None)
                sanity_check_ip_host_vs_subnet(args, cli)

            # translate 'pim' to 'ip pim'
            elif context_node.value.value in ('pim', 'msdp'):

                # "net add pim vrf BLUE rp 1.1.1.1" must configure
                #
                # vrf BLUE
                #  ip pim rp 1.1.1.1
                #
                if args.get('vrf'):
                    contexts.append("vrf %s" % args.get('<text>'))
                    del cli.match[4]
                    del cli.match[3]
                else:
                    contexts.append(None)

                ip_word = Word('ip')
                ip_word.value = 'ip'
                ip_word.value_raw = 'ip'
                ip_node = Node('ip')
                ip_node.value = ip_word
                cli.match.insert(2, ip_node)
                node_cmd_start = 2

            elif context_node.value.value == 'bgp':
                bgp_vrf = None
                afi_idx = 3
                safi_idx = 4

                if args.get('vrf'):
                    bgp_vrf = args.get('<text>')
                    node_cmd_start += 2
                    afi_idx += 2
                    safi_idx += 2

                if args.get('autonomous-system'):
                    bgp_asn = args.get('<1-4294967295>')
                else:
                    bgp_asn = self.get_bgp_asn(bgp_vrf)

                if not bgp_asn:
                    print "ERROR: You must first enable BGP via 'net add bgp autonomous-system NUMBER'."
                    return False

                if bgp_vrf:
                    contexts.append('router bgp %d vrf %s' % (bgp_asn, bgp_vrf))
                else:
                    contexts.append('router bgp %d' % bgp_asn)

                # find the subcontext for bgp
                if str(cli.match[node_cmd_start]) == '[ipv4|ipv6]':
                    ipv4_ipv6_node = cli.match[afi_idx]
                    uni_multi_node = cli.match[safi_idx]

                    if ipv4_ipv6_node.value and ipv4_ipv6_node.value.text:
                        afi = ipv4_ipv6_node.value.text
                    else:
                        afi = 'ipv4'

                    if uni_multi_node.value and uni_multi_node.value.text:
                        safi = uni_multi_node.value.text
                    else:
                        safi = 'unicast'

                    subcontext = "address-family %s %s" % (afi, safi)
                    node_cmd_start += 2

                elif str(cli.match[afi_idx]) == '[ipv4]' and str(cli.match[safi_idx]) == '[unicast]':
                    subcontext = "address-family ipv4 unicast"
                    node_cmd_start += 2

                elif str(cli.match[afi_idx]) == '[ipv6]' and str(cli.match[safi_idx]) == '[unicast]':
                    subcontext = "address-family ipv6 unicast"
                    node_cmd_start += 2

                elif str(cli.match[afi_idx]) == 'ipv4' and cli.match[safi_idx].value and cli.match[safi_idx].value.text == 'unicast':
                    subcontext = "address-family ipv4 unicast"
                    node_cmd_start += 2

                elif str(cli.match[afi_idx]) == 'ipv6' and str(cli.match[safi_idx]) == 'unicast':
                    subcontext = "address-family ipv6 unicast"
                    node_cmd_start += 2

                elif str(cli.match[afi_idx]) == 'ipv4' and cli.match[safi_idx].value and cli.match[safi_idx].value.text == 'labeled-unicast':
                    subcontext = "address-family ipv4 labeled-unicast"
                    node_cmd_start += 2

                elif str(cli.match[afi_idx]) == 'ipv6' and cli.match[safi_idx].value and cli.match[safi_idx].value.text == 'labeled-unicast':
                    subcontext = "address-family ipv6 labeled-unicast"
                    node_cmd_start += 2

                elif str(cli.match[afi_idx]) == '[l2vpn]' and str(cli.match[safi_idx]) == 'evpn':
                    subcontext = "address-family l2vpn evpn"
                    node_cmd_start += 2

                elif str(cli.match[afi_idx]) == 'evpn':
                    subcontext = "address-family l2vpn evpn"
                    node_cmd_start += 1

                elif args.get('network') and args.get('<ipv6/prefixlen>'):
                    subcontext = "address-family ipv6 unicast"

                # The "ipv4 unicast" is optional (just to make v4 easier to configure)
                # for any AFI/SAFI specific command. If they left off the "ipv4 unicast"
                # add it for them.
                if subcontext is None:
                    for x in ('aggregate-address',
                              'network',
                              'redistribute',
                              'maximum-paths',
                              'activate',
                              'allowas-in',
                              'default-originate',
                              'filter-list',
                              'next-hop-self',
                              'prefix-list',
                              'remove-private-AS',
                              'route-map',
                              'route-reflector-client',
                              'send-community',
                              'soft-reconfiguration'):
                        if x in cli.argv_expanded:
                            if x == 'route-map':
                                if 'delay-timer' not in cli.argv_expanded:
                                    subcontext = "address-family ipv4 unicast"
                                    break
                            else:
                                subcontext = "address-family ipv4 unicast"
                                break

            elif context_node.value.text == 'ospf':
                contexts.append('router ospf')

            elif context_node.value.text == 'ospf6':
                contexts.append('router ospf6')

            elif context_node.value.text == 'mpls' and args.get('ldp'):
                contexts.append('mpls ldp')
                node_cmd_start += 1

                if args.get('address-family'):
                    if args.get('ipv4'):
                        subcontext = "address-family ipv4"
                    elif args.get('ipv6'):
                        subcontext = "address-family ipv6"
                    node_cmd_start += 2

            elif context_node.value.text in ('bond', 'interface', 'vlan', 'loopback'):
                if ifnames:
                    for ifname in ifnames:
                        contexts.append('interface %s' % ifname)
                else:
                    contexts.append('interface %s' % cli.match[3].value.text)

                node_cmd_start += 1

            # global config level
            else:
                node_cmd_start = 2
                contexts.append(None)

            if cmd_is_CONTEXT:
                cmd = 'CONTEXT'
            else:
                cmd = []
                for node in cli.match[node_cmd_start:]:
                    if node.value and node.value.text:
                        if node.value.text == '<interface>':
                            cmd.append('<interface>')
                        elif node.value.text == '<bgppeer>':
                            cmd.append('<bgppeer>')
                        else:
                            cmd.append(node.value.value)

                cmd = ' '.join(map(str, cmd)).strip()

            if contexts and contexts[0]:

                # There are about two dozen commands that are in the global bgp
                # context and start with the keyword 'bgp'. 'bgp router-id x.x.x.x'
                # is one example.  We did not want the parser to read:
                #
                #    frr add bgp bgp router-id x.x.x.x
                #
                # Because the double bgp just looks/feels weird. Prepend the needed
                # bgp keyword for any of these two dozen commands.
                if contexts[0].startswith('router bgp'):
                    for bgp_bgp_cmd in bgp_bgp:
                        if cmd.startswith(bgp_bgp_cmd):
                            cmd = 'bgp ' + cmd
                            break

                    if cmd.startswith('timers'):
                        cmd = cmd.replace('timers', 'timers bgp')

                    # 'bgp default ipv4-unicast' is on by default so if the user
                    # does a 'del' we need to convert that to an 'add' of 'no bgp default ipv4-unicast'
                    if cmd == 'bgp default ipv4-unicast':
                        cmd = 'no ' + cmd

                        if args.get('del'):
                            args['del'] = None
                            args['add'] = True
                        elif args.get('add'):
                            args['del'] = True
                            args['add'] = None

                    if re.search('neighbor \S+ activate', cmd):
                        # 'ipv4 unicast' is active by default unless 'no bgp default ipv4-unicast' is configured
                        if subcontext == 'address-family ipv4 unicast':

                            # If 'no bgp default ipv4-unicast' is configured then treat this like any other address-family
                            # If not though then we must prepend the "no " string and flip the add to a del and vice versa
                            if not self.look_for_line('no bgp default ipv4-unicast', True):
                                cmd = 'no ' + cmd

                                if args.get('del'):
                                    args['del'] = None
                                    args['add'] = True
                                elif args.get('add'):
                                    args['del'] = True
                                    args['add'] = None

                    # If you configure something like "timers 2 3" frr will
                    # silently change this to "1 3" so enforce it here so the
                    # user is aware of the restriction.
                    #
                    #     net add bgp neighbor <bgppeer> timers <0-65535> <3-65535>
                    #
                    if args.get('add') and args.get('neighbor') and args.get('timers') and args.get('<3-65535>'):
                        keepalive = args.get('<0-65535>')
                        holdtime = args.get('<3-65535>')
                        if keepalive != holdtime/3:
                            print "ERROR: The BGP keepalive interval must be 1/3 the holdtime"
                            command_ok = False
                            return command_ok

                # Same problem for some ospf commands
                elif contexts[0] == 'router ospf' or contexts[0].startswith('router ospf '):
                    for ospf_ospf_cmd in ospf_ospf:
                        if cmd.startswith(ospf_ospf_cmd):
                            cmd = 'ospf ' + cmd
                            break

                # Like 'bgp default ipv4-unicast', 'ipv6 nd suppress-ra' and 'ptm-enable'
                # are on by default so we have to jump through some hoops to add the "no" line
                elif contexts[0].startswith('interface'):
                    if cmd in ('ipv6 nd suppress-ra', 'ptm-enable'):
                        cmd = 'no ' + cmd

                        if args.get('del'):
                            args['del'] = None
                            args['add'] = True
                        elif args.get('add'):
                            args['del'] = True
                            args['add'] = None

                    elif cmd.startswith('ospf6'):
                        cmd = cmd.replace('ospf6', 'ipv6 ospf6')

                    elif cmd.startswith('ospf'):
                        cmd = cmd.replace('ospf', 'ip ospf')

                    elif cmd.startswith('igmp'):
                        cmd = cmd.replace('igmp', 'ip igmp')

                    elif cmd.startswith('pim'):
                        cmd = cmd.replace('pim', 'ip pim')

                    elif cmd.startswith('msdp'):
                        cmd = cmd.replace('msdp', 'ip msdp')

            # If the user is doing a del there are several options where we can
            # del based on the keyword and do not require the value.  For example
            # the user can delete the bgp router-id using either of the following:
            #
            #   net del bgp router-id 1.1.1.1
            #   net del bgp router-id
            #
            # exact_match determines if delete() looks for the line 'bgp router-id 1.1.1.1'
            # or just look for a line that begins with 'bgp router-id'
            if str(add_del_node) == '(add|del)' or str(add_del_node) == 'add':
                exact_match = True

            elif str(add_del_node) == 'del':
                last_node = cli.match[-1]

                # If the user specified the optional arg at the end then
                # exact_match must be True
                if not last_node.value:
                    exact_match = False
                else:
                    exact_match = True

            else:
                raise Exception("Invalid add_del_node %s" % add_del_node)

            if ifnames:
                # If the user entered a glob for interfaces those need to be
                # expanded in one of two places. The first scenario is where
                # the interface name is part of the context.  For example:
                #
                #    net add inter swp3-4 ip ospf network point-to-point
                #
                # needs to add this ospf command under contexts 'interface swp3
                # and 'interface swp4'. But if the user did
                #
                #    net add bgp neighbor swp3-4 interface peer-group FOO
                #
                # we need to expand this as separate commands within the 'router bgp' context.
                #
                # So check the len of contexts to figure out which way we
                # should expand the ifnames
                if len(contexts) > 1:
                    for context in contexts:
                        if args.get('add'):
                            self.add(context, subcontext, cmd)
                        elif args.get('del'):
                            self.delete(context, subcontext, cmd, exact_match)
                else:
                    context = contexts[0]

                    for ifname in ifnames:
                        cmd_with_ifname = cmd.replace('<interface>', ifname)

                        if args.get('add'):
                            self.add(context, subcontext, cmd_with_ifname)
                        elif args.get('del'):
                            self.delete(context, subcontext, cmd_with_ifname, exact_match)
            elif bgppeers:
                context = contexts[0]

                for peer in bgppeers:
                    cmd_with_peer = cmd.replace('<bgppeer>', peer)

                    if args.get('add'):
                        self.add(context, subcontext, cmd_with_peer)
                    elif args.get('del'):
                        self.delete(context, subcontext, cmd_with_peer, exact_match)
            else:
                context = contexts[0]

                if args.get('add'):
                    self.add(context, subcontext, cmd)
                elif args.get('del'):
                    self.delete(context, subcontext, cmd, exact_match)

        else:
            print "ERROR: no commands matched"
            command_ok = False

        return command_ok


def get_config_lines(target):
    final = []

    for filename in (private_pending_conf, etc_frr_frr_conf):
        if os.path.exists(filename):
            with open(filename, 'r') as fh:
                for line in fh.readlines():
                    line = line.strip()

                    if line.startswith(target):
                        final.append(line)

    return final


def get_aspath_access_list():
    final = []

    for line in get_config_lines('ip as-path access-list'):
        # ip as-path access-list LARRY permit 1234
        final.append(line.split()[3])

    final = sort_for_humans(list(set(final)))
    return final


def get_bgp_peergroups(args={}):
    final = []
    vrf = None

    if args.get('vrf'):
        vrf = "vrf %s" % args.get('<text>')

    for filename in (private_pending_conf, etc_frr_frr_conf):
        if os.path.exists(filename):
            with open(filename, 'r') as fh:
                state = None

                for line in fh.readlines():
                    line = line.strip()

                    if line.startswith('router bgp'):
                        if vrf:
                            if line.endswith(vrf):
                                state = 'ROUTER_BGP'
                            else:
                                state = None
                        else:
                            state = 'ROUTER_BGP'

                    if line.startswith('neighbor ') and line.endswith('peer-group'):
                        if state == 'ROUTER_BGP':
                            final.append(line.split()[1])

    final = sort_for_humans(list(set(final)))
    return final


def get_bgppeers(args={}):
    final = []
    vrf = None

    if args.get('vrf'):
        vrf = "vrf %s" % args.get('<text>')

        # If the user specified 'vrf all' treat this as if they didn't
        # specify any vrf at all so that we find all peers across all vrfs
        if vrf == 'vrf all':
            vrf = None

    for filename in (private_pending_conf, etc_frr_frr_conf):
        if os.path.exists(filename):
            with open(filename, 'r') as fh:
                state = None

                for line in fh.readlines():
                    line = line.strip()

                    if line.startswith('router bgp'):
                        if vrf:
                            if line.endswith(vrf):
                                state = 'ROUTER_BGP'
                            else:
                                state = None
                        else:
                            state = 'ROUTER_BGP'

                    if line.startswith('neighbor '):
                        if state == 'ROUTER_BGP':
                            final.append(line.split()[1])

    final = sort_for_humans(list(set(final)))
    return final


def get_community_list():
    final = []

    for line in get_config_lines('ip community-list'):
        # ip community-list standard MARGE permit 100:100
        final.append(line.split()[3])

    final = sort_for_humans(list(set(final)))
    return final


def get_extcommunity_list():
    final = []

    for line in get_config_lines('ip extcommunity-list'):
        # ip community-list standard MARGE permit 100:100
        final.append(line.split()[3])

    final = sort_for_humans(list(set(final)))
    return final


def get_prefix_list_v4():
    final = []

    for line in get_config_lines('ip prefix-list'):
        # ip prefix-list BOB seq 5 permit any
        final.append(line.split()[2])

    final = sort_for_humans(list(set(final)))
    return final


def get_prefix_list_v6():
    final = []

    for line in get_config_lines('ipv6 prefix-list'):
        # ipv6 prefix-list JASON seq 5 permit any
        final.append(line.split()[2])

    final = sort_for_humans(list(set(final)))
    return final


def get_route_map():
    final = set([])

    for line in get_config_lines('route-map'):
        # route-map FOO permit 10
        final.add(line.split()[1])

    return sorted(final)


def match_aspath_access_list(argv_word, use_fuzzy, tab_completing):
    """
    <as-path-access-list>
    """
    if use_fuzzy:
        if ' ' not in argv_word:
            return True
    else:
        if argv_word in get_aspath_access_list():
            return True

        if tab_completing:
            for acl in get_aspath_access_list():
                if acl.startswith(argv_word):
                    return True
    return False


def match_bgppeer(argv_word, use_fuzzy, tab_completing):
    """
    <bgppeer>

    argv_word could be an IPv4 address, IPv6 address, ifname or glob of ifnames
    """
    if use_fuzzy:
        if ' ' not in argv_word:
            return True
    else:
        potential_peers = []

        if ifname_is_glob(str(argv_word)):
            for x in ifname_expand_glob(str(argv_word)):
                potential_peers.append(x)
        else:
            potential_peers.append(argv_word)

        all_are_peers = True
        bgp_peers = get_bgppeers()

        for peer in potential_peers:
            if peer not in bgp_peers:
                all_are_peers = False

        if all_are_peers:
            return True

        if tab_completing and len(potential_peers) == 1:
            for peer in bgp_peers:
                if peer.startswith(argv_word):
                    return True
    return False


def match_community(argv_word, use_fuzzy, tab_completing):
    """
    <community>
    There is no fuzzy matching for communities
    """
    if re.search('^\d+:\d+$', argv_word):
        return True
    return False


def match_community_list(argv_word, use_fuzzy, tab_completing):
    """
    <community-list>
    """
    if use_fuzzy:
        if ' ' not in argv_word:
            return True
    else:
        if argv_word in get_community_list():
            return True

        if tab_completing:
            for commlist in get_community_list():
                if commlist.startswith(argv_word):
                    return True
    return False


def match_extcommunity(argv_word, use_fuzzy, tab_completing):
    """
    There is no fuzzy matching for extended communities
    ASN:nn_or_IP-address:nn
    """
    if re.search('^\d+:\d+$', argv_word):
        return True

    if re.search('^\d+\.\d+\.\d+\.\d+:\d+$', argv_word):
        return True

    return False


def match_extcommunity_list(argv_word, use_fuzzy, tab_completing):
    """
    <extcommunity-list>
    """
    if use_fuzzy:
        if ' ' not in argv_word:
            return True
    else:
        if argv_word in get_extcommunity_list():
            return True

        if tab_completing:
            for extcommlist in get_extcommunity_list():
                if extcommlist.startswith(argv_word):
                    return True
    return False


def match_route_map(argv_word, use_fuzzy, tab_completing):
    """
    There is no fuzzy matching for route-maps
    """
    if use_fuzzy:
        if ' ' not in argv_word:
            return True
    else:
        if argv_word in get_route_map():
            return True

        if tab_completing:
            for routemap in get_route_map():
                if routemap.startswith(argv_word):
                    return True

    return False


def match_peer_group(argv_word, use_fuzzy, tab_completing):
    if ' ' not in argv_word:
        return True
    return False


def match_prefix_list_v4(argv_word, use_fuzzy, tab_completing):
    if use_fuzzy:
        if ' ' not in argv_word:
            return True
    else:
        if argv_word in get_prefix_list_v4():
            return True

        if tab_completing:
            for prefix_list in get_prefix_list_v4():
                if prefix_list.startswith(argv_word):
                    return True
    return False


def match_prefix_list_v6(argv_word, use_fuzzy, tab_completing):
    if use_fuzzy:
        if ' ' not in argv_word:
            return True
    else:
        if argv_word in get_prefix_list_v6():
            return True

        if tab_completing:
            for prefix_list in get_prefix_list_v6():
                if prefix_list.startswith(argv_word):
                    return True
    return False


def match_rd(argv_word, use_fuzzy, tab_completing):
    """
    <rd>
    """
    if ":" in argv_word:
        (part1, part2) = argv_word.split(':')

        if part1.isdigit() and part2.isdigit():
            part1 = int(part1)
            part2 = int(part2)

            # 4-byte:2-byte
            if part1 >= 0 and part1 <= 4294967295 and part2 >= 0 and part2 <= 65535:
                return True

            # 2-byte:4-byte
            if part1 >= 0 and part1 <= 65535 and part2 >= 0 and part2 <= 4294967295:
                return True

        # Hideous regex for matching an IPv4 address
        elif re.search('(([01]?[0-9]?[0-9]|2[0-4][0-9]|2[5][0-5])\.){3}([01]?[0-9]?[0-9]|2[0-4][0-9]|2[5][0-5])', part1):
            if part2.isdigit():
                part2 = int(part2)

                if part2 >= 0 and part2 <= 65535:
                    return True

                if use_fuzzy:
                    return True

    return False


def match_route_target(argv_word, use_fuzzy, tab_completing):
    """
    <route-target>
    """
    if ":" in argv_word:
        (part1, part2) = argv_word.split(':')

        if part1.isdigit() and part2.isdigit():
            part1 = int(part1)
            part2 = int(part2)

            # 4-byte:2-byte
            if part1 >= 0 and part1 <= 4294967295 and part2 >= 0 and part2 <= 65535:
                return True

            # 2-byte:4-byte
            if part1 >= 0 and part1 <= 65535 and part2 >= 0 and part2 <= 4294967295:
                return True

    return False


def tab_aspath_access_list(args, ended_with_space, last_argv):
    final = []
    for x in get_aspath_access_list():
        if ended_with_space or x.startswith(last_argv):
            final.append((x, 'as-path access-list'))

    if not final:
        final.append(('<as-path-access-list>', 'as-path access-list'))

    return final


def tab_bgppeer(args, ended_with_space, last_argv):
    final = []
    for x in get_bgppeers(args):
        if ended_with_space or x.startswith(last_argv):
            final.append((x, 'BGP neighbor or peer-group'))

    if not final:
        final.append(('<bgppeer>', 'BGP neighbor or peer-group'))

    return final


def tab_community_list(args, ended_with_space, last_argv):
    final = []
    for x in get_community_list():
        if ended_with_space or x.startswith(last_argv):
            final.append((x, 'BGP community-list'))

    if not final:
        final.append(('<community-list>', 'BGP community-list'))

    return final


def tab_extcommunity_list(args, ended_with_space, last_argv):
    final = []
    for x in get_extcommunity_list():
        if ended_with_space or x.startswith(last_argv):
            final.append((x, 'BGP extcommunity-list'))

    if not final:
        final.append(('<extcommunity-list>', 'BGP extcommunity-list'))

    return final


def tab_peergroup(args, ended_with_space, last_argv):
    final = []
    for x in get_bgp_peergroups(args):
        if ended_with_space or x.startswith(last_argv):
            final.append((x, 'BGP peer-group'))

    if not final:
        final.append(('<peergroup>', 'BGP peer-group'))

    return final


def tab_prefix_list_v4(args, ended_with_space, last_argv):
    final = []
    for x in get_prefix_list_v4():
        if ended_with_space or x.startswith(last_argv):
            final.append((x, 'prefix-list'))

    if not final:
        final.append(('<prefix-list>', 'prefix-list'))

    return final


def tab_prefix_list_v6(args, ended_with_space, last_argv):
    final = []
    for x in get_prefix_list_v6():
        if ended_with_space or x.startswith(last_argv):
            final.append((x, 'prefix-list'))

    if not final:
        final.append(('<prefix-list>', 'prefix-list'))

    return final


def tab_route_map(args, ended_with_space, last_argv):
    final = []

    # If ended_with_space is True, the user did something like "net show route-map <TAB>", and
    # we need to list all route-maps (in that example).  If ended_with_space is False, they did
    # "net show route-map B<TAB>", and we need to list only the route-maps that begin with 'B'.
    for x in get_route_map():
        if ended_with_space or x.startswith(last_argv):
            final.append((x, 'route-map'))

    if not final:
        final.append(('<route-map>', 'route-map'))

    return final
